# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Tests For JsonFilter.
"""

from oslo_serialization import jsonutils

from manila.scheduler.filters import json
from manila import test
from manila.tests.scheduler import fakes


class HostFiltersTestCase(test.TestCase):
    """Test case for JsonFilter."""

    def setUp(self):
        super(HostFiltersTestCase, self).setUp()
        self.json_query = jsonutils.dumps(
            ['and', ['>=', '$free_ram_mb', 1024],
             ['>=', '$free_disk_mb', 200 * 1024]])
        self.filter = json.JsonFilter()

    def test_json_filter_passes(self):
        filter_properties = {'resource_type': {'memory_mb': 1024,
                                               'root_gb': 200,
                                               'ephemeral_gb': 0},
                             'scheduler_hints': {'query': self.json_query}}
        capabilities = {'enabled': True}
        host = fakes.FakeHostState('host1',
                                   {'free_ram_mb': 1024,
                                    'free_disk_mb': 200 * 1024,
                                    'capabilities': capabilities})
        self.assertTrue(self.filter.host_passes(host, filter_properties))

    def test_json_filter_passes_with_no_query(self):
        filter_properties = {'resource_type': {'memory_mb': 1024,
                                               'root_gb': 200,
                                               'ephemeral_gb': 0}}
        capabilities = {'enabled': True}
        host = fakes.FakeHostState('host1',
                                   {'free_ram_mb': 0,
                                    'free_disk_mb': 0,
                                    'capabilities': capabilities})
        self.assertTrue(self.filter.host_passes(host, filter_properties))

    def test_json_filter_fails_on_memory(self):
        filter_properties = {'resource_type': {'memory_mb': 1024,
                                               'root_gb': 200,
                                               'ephemeral_gb': 0},
                             'scheduler_hints': {'query': self.json_query}}
        capabilities = {'enabled': True}
        host = fakes.FakeHostState('host1',
                                   {'free_ram_mb': 1023,
                                    'free_disk_mb': 200 * 1024,
                                    'capabilities': capabilities})
        self.assertFalse(self.filter.host_passes(host, filter_properties))

    def test_json_filter_fails_on_disk(self):
        filter_properties = {'resource_type': {'memory_mb': 1024,
                                               'root_gb': 200,
                                               'ephemeral_gb': 0},
                             'scheduler_hints': {'query': self.json_query}}
        capabilities = {'enabled': True}
        host = fakes.FakeHostState('host1',
                                   {'free_ram_mb': 1024,
                                    'free_disk_mb': (200 * 1024) - 1,
                                    'capabilities': capabilities})
        self.assertFalse(self.filter.host_passes(host, filter_properties))

    def test_json_filter_fails_on_caps_disabled(self):
        json_query = jsonutils.dumps(
            ['and', ['>=', '$free_ram_mb', 1024],
             ['>=', '$free_disk_mb', 200 * 1024],
             '$capabilities.enabled'])
        filter_properties = {'resource_type': {'memory_mb': 1024,
                                               'root_gb': 200,
                                               'ephemeral_gb': 0},
                             'scheduler_hints': {'query': json_query}}
        capabilities = {'enabled': False}
        host = fakes.FakeHostState('host1',
                                   {'free_ram_mb': 1024,
                                    'free_disk_mb': 200 * 1024,
                                    'capabilities': capabilities})
        self.assertFalse(self.filter.host_passes(host, filter_properties))

    def test_json_filter_fails_on_service_disabled(self):
        json_query = jsonutils.dumps(
            ['and', ['>=', '$free_ram_mb', 1024],
             ['>=', '$free_disk_mb', 200 * 1024],
             ['not', '$service.disabled']])
        filter_properties = {'resource_type': {'memory_mb': 1024,
                                               'local_gb': 200},
                             'scheduler_hints': {'query': json_query}}
        capabilities = {'enabled': True}
        host = fakes.FakeHostState('host1',
                                   {'free_ram_mb': 1024,
                                    'free_disk_mb': 200 * 1024,
                                    'capabilities': capabilities})
        self.assertFalse(self.filter.host_passes(host, filter_properties))

    def test_json_filter_happy_day(self):
        """Test json filter more thoroughly."""
        raw = ['and',
               '$capabilities.enabled',
               ['=', '$capabilities.opt1', 'match'],
               ['or',
                ['and',
                 ['<', '$free_ram_mb', 30],
                 ['<', '$free_disk_mb', 300]],
                ['and',
                 ['>', '$free_ram_mb', 30],
                 ['>', '$free_disk_mb', 300]]]]
        filter_properties = {
            'scheduler_hints': {
                'query': jsonutils.dumps(raw),
            },
        }

        # Passes
        capabilities = {'enabled': True, 'opt1': 'match'}
        service = {'disabled': False}
        host = fakes.FakeHostState('host1',
                                   {'free_ram_mb': 10,
                                    'free_disk_mb': 200,
                                    'capabilities': capabilities,
                                    'service': service})
        self.assertTrue(self.filter.host_passes(host, filter_properties))

        # Passes
        capabilities = {'enabled': True, 'opt1': 'match'}
        service = {'disabled': False}
        host = fakes.FakeHostState('host1',
                                   {'free_ram_mb': 40,
                                    'free_disk_mb': 400,
                                    'capabilities': capabilities,
                                    'service': service})
        self.assertTrue(self.filter.host_passes(host, filter_properties))

        # Fails due to capabilities being disabled
        capabilities = {'enabled': False, 'opt1': 'match'}
        service = {'disabled': False}
        host = fakes.FakeHostState('host1',
                                   {'free_ram_mb': 40,
                                    'free_disk_mb': 400,
                                    'capabilities': capabilities,
                                    'service': service})
        self.assertFalse(self.filter.host_passes(host, filter_properties))

        # Fails due to being exact memory/disk we don't want
        capabilities = {'enabled': True, 'opt1': 'match'}
        service = {'disabled': False}
        host = fakes.FakeHostState('host1',
                                   {'free_ram_mb': 30,
                                    'free_disk_mb': 300,
                                    'capabilities': capabilities,
                                    'service': service})
        self.assertFalse(self.filter.host_passes(host, filter_properties))

        # Fails due to memory lower but disk higher
        capabilities = {'enabled': True, 'opt1': 'match'}
        service = {'disabled': False}
        host = fakes.FakeHostState('host1',
                                   {'free_ram_mb': 20,
                                    'free_disk_mb': 400,
                                    'capabilities': capabilities,
                                    'service': service})
        self.assertFalse(self.filter.host_passes(host, filter_properties))

        # Fails due to capabilities 'opt1' not equal
        capabilities = {'enabled': True, 'opt1': 'no-match'}
        service = {'enabled': True}
        host = fakes.FakeHostState('host1',
                                   {'free_ram_mb': 20,
                                    'free_disk_mb': 400,
                                    'capabilities': capabilities,
                                    'service': service})
        self.assertFalse(self.filter.host_passes(host, filter_properties))

    def test_json_filter_basic_operators(self):
        host = fakes.FakeHostState('host1',
                                   {'capabilities': {'enabled': True}})
        # (operator, arguments, expected_result)
        ops_to_test = [
            ['=', [1, 1], True],
            ['=', [1, 2], False],
            ['<', [1, 2], True],
            ['<', [1, 1], False],
            ['<', [2, 1], False],
            ['>', [2, 1], True],
            ['>', [2, 2], False],
            ['>', [2, 3], False],
            ['<=', [1, 2], True],
            ['<=', [1, 1], True],
            ['<=', [2, 1], False],
            ['>=', [2, 1], True],
            ['>=', [2, 2], True],
            ['>=', [2, 3], False],
            ['in', [1, 1], True],
            ['in', [1, 1, 2, 3], True],
            ['in', [4, 1, 2, 3], False],
            ['not', [True], False],
            ['not', [False], True],
            ['or', [True, False], True],
            ['or', [False, False], False],
            ['and', [True, True], True],
            ['and', [False, False], False],
            ['and', [True, False], False],
            # Nested ((True or False) and (2 > 1)) == Passes
            ['and', [['or', True, False], ['>', 2, 1]], True]]

        for (op, args, expected) in ops_to_test:
            raw = [op] + args
            filter_properties = {
                'scheduler_hints': {
                    'query': jsonutils.dumps(raw),
                },
            }
            self.assertEqual(expected,
                             self.filter.host_passes(host, filter_properties))

        # This results in [False, True, False, True] and if any are True
        # then it passes...
        raw = ['not', True, False, True, False]
        filter_properties = {
            'scheduler_hints': {
                'query': jsonutils.dumps(raw),
            },
        }
        self.assertTrue(self.filter.host_passes(host, filter_properties))

        # This results in [False, False, False] and if any are True
        # then it passes...which this doesn't
        raw = ['not', True, True, True]
        filter_properties = {
            'scheduler_hints': {
                'query': jsonutils.dumps(raw),
            },
        }
        self.assertFalse(self.filter.host_passes(host, filter_properties))

    def test_json_filter_unknown_operator_raises(self):
        raw = ['!=', 1, 2]
        filter_properties = {
            'scheduler_hints': {
                'query': jsonutils.dumps(raw),
            },
        }
        host = fakes.FakeHostState('host1',
                                   {'capabilities': {'enabled': True}})
        self.assertRaises(KeyError,
                          self.filter.host_passes, host, filter_properties)

    def test_json_filter_empty_filters_pass(self):
        host = fakes.FakeHostState('host1',
                                   {'capabilities': {'enabled': True}})

        raw = []
        filter_properties = {
            'scheduler_hints': {
                'query': jsonutils.dumps(raw),
            },
        }
        self.assertTrue(self.filter.host_passes(host, filter_properties))
        raw = {}
        filter_properties = {
            'scheduler_hints': {
                'query': jsonutils.dumps(raw),
            },
        }
        self.assertTrue(self.filter.host_passes(host, filter_properties))

    def test_json_filter_invalid_num_arguments_fails(self):
        host = fakes.FakeHostState('host1',
                                   {'capabilities': {'enabled': True}})

        raw = ['>', ['and', ['or', ['not', ['<', ['>=', ['<=', ['in', ]]]]]]]]
        filter_properties = {
            'scheduler_hints': {
                'query': jsonutils.dumps(raw),
            },
        }
        self.assertFalse(self.filter.host_passes(host, filter_properties))

        raw = ['>', 1]
        filter_properties = {
            'scheduler_hints': {
                'query': jsonutils.dumps(raw),
            },
        }
        self.assertFalse(self.filter.host_passes(host, filter_properties))

    def test_json_filter_unknown_variable_ignored(self):
        host = fakes.FakeHostState('host1',
                                   {'capabilities': {'enabled': True}})

        raw = ['=', '$........', 1, 1]
        filter_properties = {
            'scheduler_hints': {
                'query': jsonutils.dumps(raw),
            },
        }
        self.assertTrue(self.filter.host_passes(host, filter_properties))

        raw = ['=', '$foo', 2, 2]
        filter_properties = {
            'scheduler_hints': {
                'query': jsonutils.dumps(raw),
            },
        }
        self.assertTrue(self.filter.host_passes(host, filter_properties))



# Copyright 2016 Nexenta Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re
import six

from oslo_utils import units


def str2size(s, scale=1024):
    """Convert size-string.

    String format: <value>[:space:]<B | K | M | ...> to bytes.

    :param s: size-string
    :param scale: base size
    """
    if not s:
        return 0
    if isinstance(s, six.integer_types):
        return s

    match = re.match(r'^([\.\d]+)\s*([BbKkMmGgTtPpEeZzYy]?)', s)
    if match is None:
        raise ValueError('Invalid value: %s' % s)
    groups = match.groups()
    value = float(groups[0])
    suffix = len(groups) > 1 and groups[1].upper() or 'B'
    types = ('B', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
    for i, t in enumerate(types):
        if suffix == t:
            return float(value * pow(scale, i))


def str2gib_size(s):
    """Covert size-string to size in gigabytes."""
    size_in_bytes = str2size(s)
    return size_in_bytes // units.Gi


def bytes_to_gb(size):
    return float(size) / units.Gi



# Copyright (c) 2012 OpenStack Foundation
# Copyright (c) 2015 Rushil Chugh
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Share-related Utilities and helpers."""

from oslo_config import cfg

from manila.common import constants
from manila.db import migration
from manila import rpc
from manila import utils

DEFAULT_POOL_NAME = '_pool0'
CONF = cfg.CONF


def extract_host(host, level='backend', use_default_pool_name=False):
    """Extract Host, Backend or Pool information from host string.

    :param host: String for host, which could include host@backend#pool info
    :param level: Indicate which level of information should be extracted
                  from host string. Level can be 'host', 'backend', 'pool',
                  or 'backend_name', default value is 'backend'
    :param use_default_pool_name: This flag specifies what to do
                              if level == 'pool' and there is no 'pool' info
                              encoded in host string.  default_pool_name=True
                              will return DEFAULT_POOL_NAME, otherwise it will
                              return None. Default value of this parameter
                              is False.
    :return: expected level of information

    For example:
        host = 'HostA@BackendB#PoolC'
        ret = extract_host(host, 'host')
        # ret is 'HostA'
        ret = extract_host(host, 'backend')
        # ret is 'HostA@BackendB'
        ret = extract_host(host, 'pool')
        # ret is 'PoolC'
        ret = extract_host(host, 'backend_name')
        # ret is 'BackendB'
        host = 'HostX@BackendY'
        ret = extract_host(host, 'pool')
        # ret is None
        ret = extract_host(host, 'pool', True)
        # ret is '_pool0'
    """
    if level == 'host':
        # Make sure pool is not included
        hst = host.split('#')[0]
        return hst.split('@')[0]
    if level == 'backend_name':
        hst = host.split('#')[0]
        return hst.split('@')[1]
    elif level == 'backend':
        return host.split('#')[0]
    elif level == 'pool':
        lst = host.split('#')
        if len(lst) == 2:
            return lst[1]
        elif use_default_pool_name is True:
            return DEFAULT_POOL_NAME
        else:
            return None


def append_host(host, pool):
    """Encode pool into host info."""
    if not host or not pool:
        return host

    new_host = "#".join([host, pool])
    return new_host


def get_active_replica(replica_list):
    """Returns the first 'active' replica in the list of replicas provided."""
    for replica in replica_list:
        if replica['replica_state'] == constants.REPLICA_STATE_ACTIVE:
            return replica


def change_rules_to_readonly(access_rules, add_rules, delete_rules):
    dict_access_rules = cast_access_object_to_dict_in_readonly(access_rules)
    dict_add_rules = cast_access_object_to_dict_in_readonly(add_rules)
    dict_delete_rules = cast_access_object_to_dict_in_readonly(delete_rules)
    return dict_access_rules, dict_add_rules, dict_delete_rules


def cast_access_object_to_dict_in_readonly(rules):
    dict_rules = []
    for rule in rules:
        dict_rules.append({
            'access_level': constants.ACCESS_LEVEL_RO,
            'access_type': rule['access_type'],
            'access_to': rule['access_to']
        })
    return dict_rules


@utils.if_notifications_enabled
def notify_about_share_usage(context, share, share_instance,
                             event_suffix, extra_usage_info=None, host=None):

    if not host:
        host = CONF.host

    if not extra_usage_info:
        extra_usage_info = {}

    usage_info = _usage_from_share(share, share_instance, **extra_usage_info)

    rpc.get_notifier("share", host).info(context, 'share.%s' % event_suffix,
                                         usage_info)


def _usage_from_share(share_ref, share_instance_ref, **extra_usage_info):

    usage_info = {
        'share_id': share_ref['id'],
        'user_id': share_ref['user_id'],
        'project_id': share_ref['project_id'],
        'snapshot_id': share_ref['snapshot_id'],
        'share_group_id': share_ref['share_group_id'],
        'size': share_ref['size'],
        'name': share_ref['display_name'],
        'description': share_ref['display_description'],
        'proto': share_ref['share_proto'],
        'is_public': share_ref['is_public'],
        'availability_zone': share_instance_ref['availability_zone'],
        'host': share_instance_ref['host'],
        'status': share_instance_ref['status'],
    }

    usage_info.update(extra_usage_info)

    return usage_info


def get_recent_db_migration_id():
    return migration.version()



# Copyright 2018 SAP SE
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import ddt
import mock

from manila.api.v1 import security_service
from manila.common import constants
from manila import context
from manila import test
from manila.tests.api import fakes


def stub_security_service(self, version, id):
    ss_dict = dict(
        id=id,
        name='security_service_%s' % str(id),
        type=constants.SECURITY_SERVICES_ALLOWED_TYPES[0],
        description='Fake Security Service Desc',
        dns_ip='1.1.1.1',
        server='fake-server',
        domain='fake-domain',
        user='fake-user',
        password='fake-password',
        status=constants.STATUS_NEW,
        share_networks=[],
        created_at=datetime.datetime(2017, 8, 24, 1, 1, 1, 1),
        updated_at=datetime.datetime(2017, 8, 24, 1, 1, 1, 1),
        project_id='fake-project'
    )
    if self.is_microversion_ge(version, '2.44'):
        ss_dict['ou'] = 'fake-ou'

    return ss_dict


@ddt.ddt
class SecurityServicesAPITest(test.TestCase):
    @ddt.data(
        ('2.0'),
        ('2.43'),
        ('2.44'),
    )
    def test_index(self, version):
        ss = [
            stub_security_service(self, version, 1),
            stub_security_service(self, version, 2),
        ]
        ctxt = context.RequestContext('admin', 'fake', True)
        request = fakes.HTTPRequest.blank('/security-services?all_tenants=1',
                                          version=version)
        request.headers['X-Openstack-Manila-Api-Version'] = version
        request.environ['manila.context'] = ctxt
        self.mock_object(security_service.db, 'security_service_get_all',
                         mock.Mock(return_value=ss))
        self.mock_object(security_service.db,
                         'share_network_get_all_by_security_service',
                         mock.Mock(return_value=[]))

        ss_controller = security_service.SecurityServiceController()

        result = ss_controller.detail(request)

        self.assertIsInstance(result, dict)
        self.assertEqual(['security_services'], list(result.keys()))
        self.assertIsInstance(result['security_services'], list)
        self.assertEqual(2, len(result['security_services']))
        self.assertIn(ss[0], result['security_services'])

        ss_keys = list(result['security_services'][0].keys())
        if self.is_microversion_ge(version, '2.44'):
            self.assertIn('ou', ss_keys)
        else:
            self.assertNotIn('ou', ss_keys)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manila.api import common


class ViewBuilder(common.ViewBuilder):
    """Model a server API response as a python dictionary."""

    _collection_name = 'share_instances'
    _collection_links = 'share_instances_links'

    _detail_version_modifiers = [
        "remove_export_locations",
        "add_access_rules_status_field",
        "add_replication_fields",
        "add_share_type_field",
        "add_cast_rules_to_readonly_field",
    ]

    def detail_list(self, request, instances):
        """Detailed view of a list of share instances."""
        return self._list_view(self.detail, request, instances)

    def detail(self, request, share_instance):
        """Detailed view of a single share instance."""
        export_locations = [e['path'] for e in share_instance.export_locations]

        instance_dict = {
            'id': share_instance.get('id'),
            'share_id': share_instance.get('share_id'),
            'availability_zone': share_instance.get('availability_zone'),
            'created_at': share_instance.get('created_at'),
            'host': share_instance.get('host'),
            'status': share_instance.get('status'),
            'share_network_id': share_instance.get('share_network_id'),
            'share_server_id': share_instance.get('share_server_id'),
            'export_location': share_instance.get('export_location'),
            'export_locations': export_locations,
        }
        self.update_versioned_resource_dict(
            request, instance_dict, share_instance)
        return {'share_instance': instance_dict}

    def _list_view(self, func, request, instances):
        """Provide a view for a list of share instances."""
        instances_list = [func(request, instance)['share_instance']
                          for instance in instances]
        instances_links = self._get_collection_links(request,
                                                     instances,
                                                     self._collection_name)
        instances_dict = {self._collection_name: instances_list}

        if instances_links:
            instances_dict[self._collection_links] = instances_links

        return instances_dict

    @common.ViewBuilder.versioned_method("2.9")
    def remove_export_locations(self, context, share_instance_dict,
                                share_instance):
        share_instance_dict.pop('export_location')
        share_instance_dict.pop('export_locations')

    @common.ViewBuilder.versioned_method("2.10")
    def add_access_rules_status_field(self, context, instance_dict,
                                      share_instance):
        instance_dict['access_rules_status'] = (
            share_instance.get('access_rules_status')
        )

    @common.ViewBuilder.versioned_method("2.11")
    def add_replication_fields(self, context, instance_dict, share_instance):
        instance_dict['replica_state'] = share_instance.get('replica_state')

    @common.ViewBuilder.versioned_method("2.22")
    def add_share_type_field(self, context, instance_dict, share_instance):
        instance_dict['share_type_id'] = share_instance.get('share_type_id')

    @common.ViewBuilder.versioned_method("2.30")
    def add_cast_rules_to_readonly_field(self, context, instance_dict,
                                         share_instance):
        instance_dict['cast_rules_to_readonly'] = share_instance.get(
            'cast_rules_to_readonly', False)



#   Copyright 2015 Mirantis inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from oslo_log import log
from six.moves import http_client
import webob
from webob import exc

from manila.api.openstack import wsgi
from manila.common import constants
from manila import exception
from manila.i18n import _
from manila import share

LOG = log.getLogger(__name__)


class ShareUnmanageMixin(object):

    @wsgi.Controller.authorize("unmanage")
    def _unmanage(self, req, id, body=None, allow_dhss_true=False):
        """Unmanage a share."""
        context = req.environ['manila.context']

        LOG.info("Unmanage share with id: %s", id, context=context)

        try:
            share = self.share_api.get(context, id)
            if share.get('has_replicas'):
                msg = _("Share %s has replicas. It cannot be unmanaged "
                        "until all replicas are removed.") % share['id']
                raise exc.HTTPConflict(explanation=msg)
            if (not allow_dhss_true and
                    share['instance'].get('share_server_id')):
                msg = _("Operation 'unmanage' is not supported for shares "
                        "that are created on top of share servers "
                        "(created with share-networks).")
                raise exc.HTTPForbidden(explanation=msg)
            elif share['status'] in constants.TRANSITIONAL_STATUSES:
                msg = _("Share with transitional state can not be unmanaged. "
                        "Share '%(s_id)s' is in '%(state)s' state.") % dict(
                            state=share['status'], s_id=share['id'])
                raise exc.HTTPForbidden(explanation=msg)
            snapshots = self.share_api.db.share_snapshot_get_all_for_share(
                context, id)
            if snapshots:
                msg = _("Share '%(s_id)s' can not be unmanaged because it has "
                        "'%(amount)s' dependent snapshot(s).") % {
                            's_id': id, 'amount': len(snapshots)}
                raise exc.HTTPForbidden(explanation=msg)
            self.share_api.unmanage(context, share)
        except exception.NotFound as e:
            raise exc.HTTPNotFound(explanation=e)
        except (exception.InvalidShare, exception.PolicyNotAuthorized) as e:
            raise exc.HTTPForbidden(explanation=e)

        return webob.Response(status_int=http_client.ACCEPTED)


class ShareUnmanageController(ShareUnmanageMixin, wsgi.Controller):
    """The Unmanage API controller for the OpenStack API."""

    resource_name = "share"

    def __init__(self, *args, **kwargs):
        super(ShareUnmanageController, self).__init__(*args, **kwargs)
        self.share_api = share.API()

    @wsgi.Controller.api_version('1.0', '2.6')
    def unmanage(self, req, id):
        return self._unmanage(req, id)


def create_resource():
    return wsgi.Resource(ShareUnmanageController())



# Copyright (c) 2015 Hitachi Data Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

import ddt
import mock
from oslo_concurrency import processutils as putils
from oslo_config import cfg
import paramiko
import six

from manila import exception
from manila.share.drivers.hitachi.hnas import ssh
from manila import test
from manila import utils as mutils


CONF = cfg.CONF

HNAS_RESULT_empty = ""

HNAS_RESULT_limits = """
Filesystem Ensure on span fake_fs:

Current capacity                             50GiB

Thin provision:                           disabled

Filesystem is confined to:                  100GiB
    (Run 'filesystem-confine')
Free space on span allows expansion to:     143GiB
    (Run 'span-expand')
Chunk size allows growth to:               1069GiB
    (This is a conservative estimate)
Largest filesystem that can be checked:  262144GiB
    (This is a hard limit)
This server model allows growth to:      262144GiB
    (Upgrade the server) """

HNAS_RESULT_expdel = """Deleting the export '/dir1' on fs 'fake_fs'...
NFS Export Delete: Export successfully deleted"""

HNAS_RESULT_vvoldel = """
Warning: Clearing dangling space trackers from empty vivol"""

HNAS_RESULT_selectfs = "Current selected file system: fake_fs, number(1)"

HNAS_RESULT_expadd = "NFS Export Add: Export added successfully"

HNAS_RESULT_vvol = """vvol_test
  email        :
  root         : /vvol_test
  tag          : 39
  usage  bytes : 0 B  files: 1
  last modified: 2015-06-23 22:36:12.830698800+00:00"""

HNAS_RESULT_vvol_error = "The virtual volume does not exist."

HNAS_RESULT_mount = """ \
Request to mount file system fake_fs submitted successfully.
File system fake_fs successfully mounted."""

HNAS_RESULT_quota = """Type            : Explicit
Target          : ViVol: vvol_test
Usage           : 1 GB
  Limit         : 5 GB (Hard)
  Warning       : Unset
  Critical      : Unset
  Reset         : 5% (51.2 MB)
File Count      : 1
  Limit         : Unset
  Warning       : Unset
  Critical      : Unset
  Reset         : 5% (0)
Generate Events : Disabled
Global id       : 28a3c9f8-ae05-11d0-9025-836896aada5d
Last modified   : 2015-06-23 22:37:17.363660800+00:00 """

HNAS_RESULT_quota_tb = """Type            : Explicit
Target          : ViVol: vvol_test
Usage           : 1 TB
  Limit         : 1 TB (Hard)
  Warning       : Unset
  Critical      : Unset
  Reset         : 5% (51.2 MB)
File Count      : 1
  Limit         : Unset
  Warning       : Unset
  Critical      : Unset
  Reset         : 5% (0)
Generate Events : Disabled
Global id       : 28a3c9f8-ae05-11d0-9025-836896aada5d
Last modified   : 2015-06-23 22:37:17.363660800+00:00  """

HNAS_RESULT_quota_mb = """Type            : Explicit
Target          : ViVol: vvol_test
Usage           : 20 MB
  Limit         : 500 MB (Hard)
  Warning       : Unset
  Critical      : Unset
  Reset         : 5% (51.2 MB)
File Count      : 1
  Limit         : Unset
  Warning       : Unset
  Critical      : Unset
  Reset         : 5% (0)
Generate Events : Disabled
Global id       : 28a3c9f8-ae05-11d0-9025-836896aada5d
Last modified   : 2015-06-23 22:37:17.363660800+00:00  """

HNAS_RESULT_quota_unset = """Type            : Explicit
Target          : ViVol: vvol_test
Usage           : 0 B
  Limit         : Unset
  Warning       : Unset
  Critical      : Unset
  Reset         : 5% (51.2 MB)
File Count      : 1
  Limit         : Unset
  Warning       : Unset
  Critical      : Unset
  Reset         : 5% (0)
Generate Events : Disabled
Global id       : 28a3c9f8-ae05-11d0-9025-836896aada5d
Last modified   : 2015-06-23 22:37:17.363660800+00:00  """

HNAS_RESULT_quota_err = """No quotas matching specified filter criteria.
"""

HNAS_RESULT_export = """Export name: vvol_test
            Export path: /vvol_test
      File system label: file_system
       File system size: 3.969 GB
 File system free space: 1.848 GB
      File system state:
               formatted = Yes
                 mounted = Yes
                  failed = No
        thin provisioned = No
       Access snapshots: No
      Display snapshots: No
           Read Caching: Disabled
Disaster recovery setting:
                Recovered = No
         Transfer setting = Use file system default \n
   Export configuration:\n
127.0.0.2
"""

HNAS_RESULT_wrong_export = """Export name: wrong_name
            Export path: /vvol_test
      File system label: file_system
       File system size: 3.969 GB
 File system free space: 1.848 GB
      File system state:
               formatted = Yes
                 mounted = Yes
                  failed = No
        thin provisioned = No
       Access snapshots: No
      Display snapshots: No
           Read Caching: Disabled
Disaster recovery setting:
                Recovered = No
         Transfer setting = Use file system default
   Export configuration:
127.0.0.1"""

HNAS_RESULT_exp_no_fs = """
            Export name: no_fs
            Export path: /export_without_fs
       File system info: *** not available ***
       Access snapshots: Yes
      Display snapshots: Yes
           Read Caching: Disabled
Disaster recovery setting:
                Recovered = No
         Transfer setting = Use file system default
   Export configuration:
   """

HNAS_RESULT_export_ip = """
            Export name: vvol_test
            Export path: /vvol_test
      File system label: fake_fs
       File system size: 3.969 GB
 File system free space: 1.848 GB
      File system state:
               formatted = Yes
                 mounted = Yes
                  failed = No
        thin provisioned = No
       Access snapshots: No
      Display snapshots: No
           Read Caching: Disabled
Disaster recovery setting:
                Recovered = No
         Transfer setting = Use file system default
   Export configuration:
127.0.0.1(rw)
"""

HNAS_RESULT_export_ip2 = """
            Export name: vvol_test
            Export path: /vvol_test
      File system label: fake_fs
       File system size: 3.969 GB
 File system free space: 1.848 GB
      File system state:
               formatted = Yes
                 mounted = Yes
                  failed = No
        thin provisioned = No
       Access snapshots: No
      Display snapshots: No
           Read Caching: Disabled
Disaster recovery setting:
                Recovered = No
         Transfer setting = Use file system default
   Export configuration:
127.0.0.1(ro)
"""

HNAS_RESULT_expmod = """Modifying the export '/fake_export' on fs 'fake_fs'...
NFS Export Modify: changing configuration options to: 127.0.0.2 NFS
Export Modify: Export modified successfully"""

HNAS_RESULT_expnotmod = "Export not modified."

HNAS_RESULT_job = """tree-operation-job-submit: Request submitted successfully.
tree-operation-job-submit: Job id = d933100a-b5f6-11d0-91d9-836896aada5d"""

HNAS_RESULT_vvol_list = """vol1
  email        :
  root         : /shares/vol1
  tag          : 10
  usage  bytes : 0 B  files: 1
  last modified: 2015-07-27 22:25:02.746426000+00:00
vol2
  email        :
  root         : /shares/vol2
  tag          : 13
  usage  bytes : 0 B  files: 1
  last modified: 2015-07-28 01:30:21.125671700+00:00
vol3
  email        :
  root         : /shares/vol3
  tag          : 14
  usage  bytes : 5 GB (5368709120 B)  files: 2
  last modified: 2015-07-28 20:23:05.672404600+00:00"""

HNAS_RESULT_tree_job_status_fail = """JOB ID : d933100a-b5f6-11d0-91d9-836896aada5d
      Job request
        Physical node                  : 1
        EVS                            : 1
        Volume number                  : 1
        File system id                 : 2ea361c20ed0f80d0000000000000000
        File system name               : fs1
        Source path                    : "/foo"
        Creation time                  : 2013-09-05 23:16:48-07:00
        Destination path               : "/clone/bar"
        Ensure destination path exists : true

      Job state                        : Job failed
      Job info
        Started                        : 2013-09-05 23:16:48-07:00
        Ended                          : 2013-09-05 23:17:02-07:00
        Status                         : Success
        Error details                  :
        Directories processed          : 220
        Files processed                : 910
        Data bytes processed           : 34.5 MB (36174754 B)
        Source directories missing     : 0
        Source files missing           : 0
        Source files skipped           : 801
        Skipping details               : 104 symlinks, 452 hard links,
47 block special devices, 25 character devices"""

HNAS_RESULT_job_completed = """JOB ID : ab4211b8-aac8-11ce-91af-39e0822ea368
      Job request
        Physical node                  : 1
        EVS                            : 1
        Volume number                  : 1
        File system id                 : 2ea361c20ed0f80d0000000000000000
        File system name               : fs1
        Source path                    : "/foo"
        Creation time                  : 2013-09-05 23:16:48-07:00
        Destination path               : "/clone/bar"
        Ensure destination path exists : true

      Job state                        : Job was completed
      Job info
        Started                        : 2013-09-05 23:16:48-07:00
        Ended                          : 2013-09-05 23:17:02-07:00
        Status                         : Success
        Error details                  :
        Directories processed          : 220
        Files processed                : 910
        Data bytes processed           : 34.5 MB (36174754 B)
        Source directories missing     : 0
        Source files missing           : 0
        Source files skipped           : 801
        Skipping details               : 104 symlinks, 452 hard links, 47 \
block special devices, 25 character devices
"""

HNAS_RESULT_job_running = """JOB ID : ab4211b8-aac8-11ce-91af-39e0822ea368
      Job request
        Physical node                  : 1
        EVS                            : 1
        Volume number                  : 1
        File system id                 : 2ea361c20ed0f80d0000000000000000
        File system name               : fs1
        Source path                    : "/foo"
        Creation time                  : 2013-09-05 23:16:48-07:00
        Destination path               : "/clone/bar"
        Ensure destination path exists : true

      Job state                        : Job is running
      Job info
        Started                        : 2013-09-05 23:16:48-07:00
        Ended                          : 2013-09-05 23:17:02-07:00
        Status                         : Success
        Error details                  :
        Directories processed          : 220
        Files processed                : 910
        Data bytes processed           : 34.5 MB (36174754 B)
        Source directories missing     : 0
        Source files missing           : 0
        Source files skipped           : 801
        Skipping details               : 104 symlinks, 452 hard links, 47 \
block special devices, 25 character devices
"""

HNAS_RESULT_df = """
  ID          Label  EVS      Size            Used  Snapshots  Deduped  \
          Avail  Thin  ThinSize  ThinAvail              FS Type
----  -------------  ---  --------  --------------  ---------  -------  \
-------------  ----  --------  ---------  -------------------
1051  FS-ManilaDev1    3  70.00 GB  10.00 GB (75%)   0 B (0%)       NA  \
18.3 GB (25%)    No                       4 KB,WFS-2,128 DSBs
"""

HNAS_RESULT_df_tb = """
  ID          Label  EVS      Size            Used  Snapshots  Deduped  \
          Avail  Thin  ThinSize  ThinAvail              FS Type
----  -------------  ---  --------  --------------  ---------  -------  \
-------------  ----  --------  ---------  -------------------
1051  FS-ManilaDev1    3.00  7.00 TB  2 TB (75%)   0 B (0%)       NA  \
18.3 GB (25%)    No                       4 KB,WFS-2,128 DSBs
"""

HNAS_RESULT_df_dedupe_on = """
  ID          Label  EVS      Size            Used  Snapshots  Deduped  \
          Avail  Thin  ThinSize  ThinAvail              FS Type
----  -------------  ---  --------  --------------  ---------  -------  \
-------------  ----  --------  ---------  -------------------
1051  FS-ManilaDev1    3.00  7.00 TB  2 TB (75%)       NA     0 B (0%)  \
18.3 GB (25%)    No                       4 KB,WFS-2,128 DSBs,dedupe enabled
"""

HNAS_RESULT_df_unmounted = """
  ID          Label  EVS      Size            Used  Snapshots  Deduped  \
          Avail  Thin  ThinSize  ThinAvail              FS Type
----  -------------  ---  --------  --------------  ---------  -------  \
-------------  ----  --------  ---------  -------------------
1051  FS-ManilaDev1    3  70.00 GB  Not mounted   0 B (0%)       NA  \
18.3 GB (25%)    No                       4 KB,WFS-2,128 DSBs
"""

HNAS_RESULT_df_error = """File system file_system not found"""

HNAS_RESULT_mounted_filesystem = """
file_system        1055  fake_span           Mount   2        4        5    1
"""

HNAS_RESULT_unmounted_filesystem = """
file_system        1055  fake_span          Umount   2        4        5    1
"""

HNAS_RESULT_cifs_list = """
           Share name: vvol_test
           Share path: \\\\shares\\vvol_test
          Share users: 2
         Share online: Yes
        Share comment:
        Cache options: Manual local caching for documents
            ABE enabled: Yes
Continuous Availability: No
       Access snapshots: No
      Display snapshots: No
     ShadowCopy enabled: Yes
   Lower case on create: No
        Follow symlinks: Yes
 Follow global symlinks: No
       Scan for viruses: Yes
     File system label: file_system
      File system size: 9.938 GB
File system free space: 6.763 GB
     File system state:
                formatted = Yes
                  mounted = Yes
                   failed = No
         thin provisioned = No
Disaster recovery setting:
                Recovered = No
         Transfer setting = Use file system default
     Home directories: Off
  Mount point options:
"""

HNAS_RESULT_different_fs_cifs_list = """
           Share name: vvol_test
           Share path: \\\\shares\\vvol_test
          Share users: 0
         Share online: Yes
        Share comment:
        Cache options: Manual local caching for documents
            ABE enabled: Yes
Continuous Availability: No
       Access snapshots: No
      Display snapshots: No
     ShadowCopy enabled: Yes
   Lower case on create: No
        Follow symlinks: Yes
 Follow global symlinks: No
       Scan for viruses: Yes
     File system label: different_filesystem
      File system size: 9.938 GB
File system free space: 6.763 GB
     File system state:
                formatted = Yes
                  mounted = Yes
                   failed = No
         thin provisioned = No
Disaster recovery setting:
                Recovered = No
         Transfer setting = Use file system default
     Home directories: Off
  Mount point options:
"""

HNAS_RESULT_list_cifs_permissions = """ \
Displaying the details of the share 'vvol_test' on file system 'filesystem' ...
Maximum user count is unlimited
Type Permission          User/Group
U    Deny  Read          NFSv4 user\\user1@domain.com
G    Deny  Change & Read Unix user\\1087
U    Allow Full Control  Unix user\\1088
U    Allow Read          Unix user\\1089
?    Deny  Full Control  NFSv4 user\\user2@company.com
X    Allow Change & Read Unix user\\1090

"""

HNAS_RESULT_check_snap_error = """ \
path-to-object-number/FS-TestCG: Unable to locate component: share1
path-to-object-number/FS-TestCG: Failed to resolve object number"""


@ddt.ddt
class HNASSSHTestCase(test.TestCase):
    def setUp(self):
        super(HNASSSHTestCase, self).setUp()

        self.ip = '192.168.1.1'
        self.port = 22
        self.user = 'hnas_user'
        self.password = 'hnas_password'
        self.default_commands = ['ssc', '127.0.0.1']
        self.fs_name = 'file_system'
        self.evs_ip = '172.24.44.1'
        self.evs_id = 2
        self.ssh_private_key = 'private_key'
        self.cluster_admin_ip0 = 'fake'
        self.job_timeout = 30

        self.mock_log = self.mock_object(ssh, 'LOG')

        self._driver_ssh = ssh.HNASSSHBackend(self.ip, self.user,
                                              self.password,
                                              self.ssh_private_key,
                                              self.cluster_admin_ip0,
                                              self.evs_id, self.evs_ip,
                                              self.fs_name, self.job_timeout)

        self.vvol = {
            'id': 'vvol_test',
            'share_proto': 'nfs',
            'size': 4,
            'host': '127.0.0.1',
        }

        self.snapshot = {
            'id': 'snapshot_test',
            'share_proto': 'nfs',
            'size': 4,
            'share_id': 'vvol_test',
            'host': 'ubuntu@hitachi2#HITACHI2',
        }
        self.mock_log.debug.reset_mock()

    def test_get_stats(self):
        fake_list_command = ['df', '-a', '-f', self.fs_name]

        self.mock_object(ssh.HNASSSHBackend, '_execute',
                         mock.Mock(return_value=(HNAS_RESULT_df_tb, "")))

        total, free, dedupe = self._driver_ssh.get_stats()

        ssh.HNASSSHBackend._execute.assert_called_with(fake_list_command)
        self.assertEqual(7168.0, total)
        self.assertEqual(5120.0, free)
        self.assertFalse(dedupe)

    def test_get_stats_dedupe_on(self):
        fake_list_command = ['df', '-a', '-f', self.fs_name]

        self.mock_object(
            ssh.HNASSSHBackend, '_execute',
            mock.Mock(return_value=(HNAS_RESULT_df_dedupe_on, "")))

        total, free, dedupe = self._driver_ssh.get_stats()

        ssh.HNASSSHBackend._execute.assert_called_with(fake_list_command)
        self.assertEqual(7168.0, total)
        self.assertEqual(5120.0, free)
        self.assertTrue(dedupe)

    def test_get_stats_error(self):

        fake_list_command = ['df', '-a', '-f', self.fs_name]

        self.mock_object(ssh.HNASSSHBackend, '_execute',
                         mock.Mock(side_effect=putils.ProcessExecutionError))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.get_stats)

        ssh.HNASSSHBackend._execute.assert_called_with(fake_list_command)

    @ddt.data(True, False)
    def test_nfs_export_add(self, is_snapshot):
        if is_snapshot:
            name = '/snapshots/fake_snap'
            path = '/snapshots/fake_share/fake_snap'
        else:
            name = path = '/shares/fake_share'

        fake_nfs_command = ['nfs-export', 'add', '-S', 'disable', '-c',
                            '127.0.0.1', name, self.fs_name,
                            path]
        self.mock_object(ssh.HNASSSHBackend, '_execute')

        if is_snapshot:
            self._driver_ssh.nfs_export_add('fake_share',
                                            snapshot_id='fake_snap')
        else:
            self._driver_ssh.nfs_export_add('fake_share')

        self._driver_ssh._execute.assert_called_with(fake_nfs_command)

    def test_nfs_export_add_error(self):
        self.mock_object(ssh.HNASSSHBackend, '_execute', mock.Mock(
            side_effect=[putils.ProcessExecutionError(stderr='')]))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.nfs_export_add, 'vvol_test')
        self.assertTrue(self.mock_log.exception.called)

    @ddt.data(True, False)
    def test_nfs_export_del(self, is_snapshot):
        if is_snapshot:
            name = '/snapshots/vvol_test'
            args = {'snapshot_id': 'vvol_test'}
        else:
            name = '/shares/vvol_test'
            args = {'share_id': 'vvol_test'}

        fake_nfs_command = ['nfs-export', 'del', name]
        self.mock_object(ssh.HNASSSHBackend, '_execute')

        self._driver_ssh.nfs_export_del(**args)

        self._driver_ssh._execute.assert_called_with(fake_nfs_command)

    def test_nfs_export_del_inexistent_export(self):
        self.mock_object(ssh.HNASSSHBackend, '_execute', mock.Mock(
            side_effect=[putils.ProcessExecutionError(
                stderr='does not exist')]))

        self._driver_ssh.nfs_export_del('vvol_test')

        self.assertTrue(self.mock_log.warning.called)

    def test_nfs_export_del_exception(self):
        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.nfs_export_del)

    def test_nfs_export_del_execute_error(self):
        self.mock_object(ssh.HNASSSHBackend, '_execute', mock.Mock(
            side_effect=[putils.ProcessExecutionError(stderr='')]))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.nfs_export_del, 'vvol_test')
        self.assertTrue(self.mock_log.exception.called)

    @ddt.data(True, False)
    def test_cifs_share_add(self, is_snapshot):
        if is_snapshot:
            name = 'fake_snap'
            path = r'\\snapshots\\fake_share\\fake_snap'
        else:
            name = 'fake_share'
            path = r'\\shares\\fake_share'

        fake_cifs_add_command = ['cifs-share', 'add', '-S', 'disable',
                                 '--enable-abe', '--nodefaultsaa',
                                 name, self.fs_name,
                                 path]
        self.mock_object(ssh.HNASSSHBackend, '_execute')

        if is_snapshot:
            self._driver_ssh.cifs_share_add('fake_share',
                                            snapshot_id='fake_snap')
        else:
            self._driver_ssh.cifs_share_add('fake_share')

        self._driver_ssh._execute.assert_called_with(fake_cifs_add_command)

    def test_cifs_share_add_error(self):
        self.mock_object(ssh.HNASSSHBackend, '_execute', mock.Mock(
            side_effect=[putils.ProcessExecutionError(stderr='')]))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.cifs_share_add, 'vvol_test')
        self.assertTrue(self.mock_log.exception.called)

    def test_cifs_share_del(self):
        fake_cifs_del_command = ['cifs-share', 'del', '--target-label',
                                 self.fs_name, 'vvol_test']
        self.mock_object(ssh.HNASSSHBackend, '_execute')

        self._driver_ssh.cifs_share_del('vvol_test')

        self._driver_ssh._execute.assert_called_with(fake_cifs_del_command)

    def test_cifs_share_del_inexistent_share(self):
        fake_cifs_del_command = ['cifs-share', 'del', '--target-label',
                                 self.fs_name, 'vvol_test']
        self.mock_object(ssh.HNASSSHBackend, '_execute',
                         mock.Mock(side_effect=putils.ProcessExecutionError(
                             exit_code=1)))

        self._driver_ssh.cifs_share_del('vvol_test')

        self._driver_ssh._execute.assert_called_with(fake_cifs_del_command)
        self.assertTrue(self.mock_log.warning.called)

    def test_cifs_share_del_exception(self):
        fake_cifs_del_command = ['cifs-share', 'del', '--target-label',
                                 self.fs_name, 'vvol_test']
        self.mock_object(ssh.HNASSSHBackend, '_execute',
                         mock.Mock(side_effect=putils.ProcessExecutionError))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.cifs_share_del, 'vvol_test')
        self._driver_ssh._execute.assert_called_with(fake_cifs_del_command)

    def test_get_nfs_host_list(self):
        self.mock_object(ssh.HNASSSHBackend, "_get_export", mock.Mock(
            return_value=[ssh.Export(HNAS_RESULT_export)]))

        host_list = self._driver_ssh.get_nfs_host_list('fake_id')

        self.assertEqual(['127.0.0.2'], host_list)

    def test_update_nfs_access_rule_empty_host_list(self):
        fake_export_command = ['nfs-export', 'mod', '-c', '127.0.0.1',
                               '/snapshots/fake_id']
        self.mock_object(ssh.HNASSSHBackend, "_execute")

        self._driver_ssh.update_nfs_access_rule([], snapshot_id="fake_id")

        self._driver_ssh._execute.assert_called_with(fake_export_command)

    def test_update_nfs_access_rule(self):
        fake_export_command = ['nfs-export', 'mod', '-c',
                               u'"127.0.0.1,127.0.0.2"', '/shares/fake_id']
        self.mock_object(ssh.HNASSSHBackend, "_execute")

        self._driver_ssh.update_nfs_access_rule(['127.0.0.1', '127.0.0.2'],
                                                share_id="fake_id")

        self._driver_ssh._execute.assert_called_with(fake_export_command)

    def test_update_nfs_access_rule_exception_no_share_provided(self):
        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.update_nfs_access_rule,
                          ['127.0.0.1'])

    def test_update_nfs_access_rule_exception_error(self):

        fake_export_command = ['nfs-export', 'mod', '-c',
                               u'"127.0.0.1,127.0.0.2"', '/shares/fake_id']
        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            side_effect=putils.ProcessExecutionError))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.update_nfs_access_rule,
                          ['127.0.0.1', '127.0.0.2'], share_id="fake_id")

        self._driver_ssh._execute.assert_called_with(fake_export_command)

    def test_cifs_allow_access(self):
        fake_cifs_allow_command = ['cifs-saa', 'add', '--target-label',
                                   self.fs_name, 'vvol_test',
                                   'fake_user', 'ar']
        self.mock_object(ssh.HNASSSHBackend, '_execute')

        self._driver_ssh.cifs_allow_access('vvol_test', 'fake_user', 'ar')

        self._driver_ssh._execute.assert_called_with(fake_cifs_allow_command)

    @ddt.data(True, False)
    def test_cifs_allow_access_already_allowed_user(self, is_snapshot):
        fake_cifs_allow_command = ['cifs-saa', 'add', '--target-label',
                                   self.fs_name, 'vvol_test',
                                   'fake_user', 'acr']
        if not is_snapshot:
            fake_cifs_allow_command2 = ['cifs-saa', 'change', '--target-label',
                                        'file_system', 'vvol_test',
                                        'fake_user', 'acr']

        self.mock_object(ssh.HNASSSHBackend, '_execute',
                         mock.Mock(side_effect=[putils.ProcessExecutionError(
                             stderr='already listed as a user'),
                             "Rule modified."]))

        self._driver_ssh.cifs_allow_access('vvol_test', 'fake_user', 'acr',
                                           is_snapshot=is_snapshot)

        _execute_calls = [mock.call(fake_cifs_allow_command)]
        if not is_snapshot:
            _execute_calls.append(mock.call(fake_cifs_allow_command2))

        self._driver_ssh._execute.assert_has_calls(_execute_calls)
        self.assertTrue(self.mock_log.debug.called)

    @ddt.data(True, False)
    def test_cifs_allow_access_exception(self, is_snapshot):
        fake_cifs_allow_command = ['cifs-saa', 'add', '--target-label',
                                   self.fs_name, 'vvol_test',
                                   'fake_user', 'acr']
        self.mock_object(ssh.HNASSSHBackend, '_execute',
                         mock.Mock(side_effect=[putils.ProcessExecutionError(
                             stderr='Could not add user/group fake_user to '
                                    'share \'vvol_test\'')]))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.cifs_allow_access, 'vvol_test',
                          'fake_user', 'acr', is_snapshot=is_snapshot)

        self._driver_ssh._execute.assert_called_with(fake_cifs_allow_command)

    def test_cifs_update_access_level_exception(self):
        fake_cifs_allow_command = ['cifs-saa', 'add', '--target-label',
                                   self.fs_name, 'vvol_test',
                                   'fake_user', 'acr']
        fake_cifs_allow_command2 = ['cifs-saa', 'change', '--target-label',
                                    'file_system', 'vvol_test', 'fake_user',
                                    'acr']

        self.mock_object(ssh.HNASSSHBackend, '_execute',
                         mock.Mock(side_effect=[putils.ProcessExecutionError(
                             stderr='already listed as a user'),
                             putils.ProcessExecutionError(
                             stderr='Error when trying to modify rule.')]))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.cifs_allow_access, 'vvol_test',
                          'fake_user', 'acr')

        self._driver_ssh._execute.assert_has_calls(
            [mock.call(fake_cifs_allow_command),
             mock.call(fake_cifs_allow_command2)])
        self.assertTrue(self.mock_log.debug.called)

    def test_cifs_deny_access(self):
        fake_cifs_deny_command = ['cifs-saa', 'delete', '--target-label',
                                  self.fs_name, 'vvol_test', 'fake_user']
        self.mock_object(ssh.HNASSSHBackend, '_execute')

        self._driver_ssh.cifs_deny_access('vvol_test', 'fake_user')

        self._driver_ssh._execute.assert_called_with(fake_cifs_deny_command)

    @ddt.data(True, False)
    def test_cifs_deny_access_already_deleted_user(self, is_snapshot):
        fake_cifs_deny_command = ['cifs-saa', 'delete', '--target-label',
                                  self.fs_name, 'vvol_test', 'fake_user']
        self.mock_object(ssh.HNASSSHBackend, '_execute', mock.Mock(
            side_effect=[putils.ProcessExecutionError(
                stderr='not listed as a user')]))

        self._driver_ssh.cifs_deny_access('vvol_test', 'fake_user',
                                          is_snapshot=is_snapshot)

        self._driver_ssh._execute.assert_called_with(fake_cifs_deny_command)
        self.assertTrue(self.mock_log.warning.called)

    def test_cifs_deny_access_backend_exception(self):
        fake_cifs_deny_command = ['cifs-saa', 'delete', '--target-label',
                                  self.fs_name, 'vvol_test', 'fake_user']
        self.mock_object(ssh.HNASSSHBackend, '_execute',
                         mock.Mock(side_effect=[putils.ProcessExecutionError(
                             stderr='Unexpected error')]))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.cifs_deny_access, 'vvol_test',
                          'fake_user')

        self._driver_ssh._execute.assert_called_with(fake_cifs_deny_command)

    def test_list_cifs_permission(self):
        fake_cifs_list_command = ['cifs-saa', 'list', '--target-label',
                                  self.fs_name, 'vvol_test']

        expected_out = ssh.CIFSPermissions(HNAS_RESULT_list_cifs_permissions)

        self.mock_object(ssh.HNASSSHBackend, '_execute', mock.Mock(
            return_value=(HNAS_RESULT_list_cifs_permissions, '')))

        out = self._driver_ssh.list_cifs_permissions('vvol_test')

        for i in range(len(expected_out.permission_list)):
            self.assertEqual(expected_out.permission_list[i], out[i])

        self._driver_ssh._execute.assert_called_with(fake_cifs_list_command)

    def test_list_cifs_no_permissions_added(self):
        fake_cifs_list_command = ['cifs-saa', 'list', '--target-label',
                                  self.fs_name, 'vvol_test']

        self.mock_object(ssh.HNASSSHBackend, '_execute', mock.Mock(
            side_effect=[putils.ProcessExecutionError(
                stderr='No entries for this share')]))

        out = self._driver_ssh.list_cifs_permissions('vvol_test')

        self.assertEqual([], out)
        self._driver_ssh._execute.assert_called_with(fake_cifs_list_command)
        self.assertTrue(self.mock_log.debug.called)

    def test_list_cifs_exception(self):
        fake_cifs_list_command = ['cifs-saa', 'list', '--target-label',
                                  self.fs_name, 'vvol_test']

        self.mock_object(ssh.HNASSSHBackend, '_execute', mock.Mock(
            side_effect=[putils.ProcessExecutionError(
                stderr='Error.')]))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.list_cifs_permissions,
                          "vvol_test")

        self._driver_ssh._execute.assert_called_with(fake_cifs_list_command)
        self.assertTrue(self.mock_log.exception.called)

    def test_tree_clone_nothing_to_clone(self):
        fake_tree_clone_command = ['tree-clone-job-submit', '-e', '-f',
                                   self.fs_name, '/src', '/dst']
        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            side_effect=[putils.ProcessExecutionError(
                stderr='Cannot find any clonable files in the source directory'
            )]))

        self.assertRaises(exception.HNASNothingToCloneException,
                          self._driver_ssh.tree_clone, "/src", "/dst")
        self._driver_ssh._execute.assert_called_with(fake_tree_clone_command)

    def test_tree_clone_error_cloning(self):
        fake_tree_clone_command = ['tree-clone-job-submit', '-e', '-f',
                                   self.fs_name, '/src', '/dst']
        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            side_effect=[putils.ProcessExecutionError(stderr='')]))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.tree_clone, "/src", "/dst")
        self._driver_ssh._execute.assert_called_with(fake_tree_clone_command)
        self.assertTrue(self.mock_log.exception.called)

    def test_tree_clone(self):
        fake_tree_clone_command = ['tree-clone-job-submit', '-e', '-f',
                                   self.fs_name, '/src', '/dst']
        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            side_effect=[(HNAS_RESULT_job, ''),
                         (HNAS_RESULT_job_completed, '')]))

        self._driver_ssh.tree_clone("/src", "/dst")

        self._driver_ssh._execute.assert_any_call(fake_tree_clone_command)
        self.assertTrue(self.mock_log.debug.called)

    def test_tree_clone_job_failed(self):
        fake_tree_clone_command = ['tree-clone-job-submit', '-e', '-f',
                                   self.fs_name, '/src', '/dst']
        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            side_effect=[(HNAS_RESULT_job, ''),
                         (HNAS_RESULT_tree_job_status_fail, '')]))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.tree_clone, "/src", "/dst")
        self._driver_ssh._execute.assert_any_call(fake_tree_clone_command)
        self.assertTrue(self.mock_log.error.called)

    def test_tree_clone_job_timeout(self):
        fake_tree_clone_command = ['tree-clone-job-submit', '-e', '-f',
                                   self.fs_name, '/src', '/dst']
        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            side_effect=[(HNAS_RESULT_job, ''),
                         (HNAS_RESULT_job_running, ''),
                         (HNAS_RESULT_job_running, ''),
                         (HNAS_RESULT_job_running, ''),
                         (HNAS_RESULT_empty, '')]))
        self.mock_object(time, "time", mock.Mock(side_effect=[0, 0, 200, 200]))
        self.mock_object(time, "sleep")

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.tree_clone, "/src", "/dst")
        self._driver_ssh._execute.assert_any_call(fake_tree_clone_command)
        self.assertTrue(self.mock_log.error.called)

    def test_tree_delete_path_does_not_exist(self):
        fake_tree_delete_command = ['tree-delete-job-submit', '--confirm',
                                    '-f', self.fs_name, '/path']
        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            side_effect=[putils.ProcessExecutionError(
                stderr='Source path: Cannot access')]
        ))

        self._driver_ssh.tree_delete("/path")

        self.assertTrue(self.mock_log.warning.called)
        self._driver_ssh._execute.assert_called_with(fake_tree_delete_command)

    def test_tree_delete_error(self):
        fake_tree_delete_command = ['tree-delete-job-submit', '--confirm',
                                    '-f', self.fs_name, '/path']
        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            side_effect=[putils.ProcessExecutionError(
                stderr='')]
        ))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.tree_delete, "/path")
        self.assertTrue(self.mock_log.exception.called)
        self._driver_ssh._execute.assert_called_with(fake_tree_delete_command)

    def test_create_directory(self):
        locked_selectfs_args = ['create', '/path']
        self.mock_object(ssh.HNASSSHBackend, "_locked_selectfs")
        self.mock_object(ssh.HNASSSHBackend, "check_directory",
                         mock.Mock(return_value=True))

        self._driver_ssh.create_directory("/path")

        self._driver_ssh._locked_selectfs.assert_called_with(
            *locked_selectfs_args)
        ssh.HNASSSHBackend.check_directory.assert_called_once_with('/path')
        self.assertFalse(self.mock_log.warning.called)

    def test_create_directory_context_change_fail(self):
        locked_selectfs_args = ['create', '/path']
        self.mock_object(time, 'sleep')
        self.mock_object(ssh.HNASSSHBackend, "_locked_selectfs")
        self.mock_object(ssh.HNASSSHBackend, "check_directory",
                         mock.Mock(return_value=False))

        self.assertRaises(exception.HNASSSCContextChange,
                          self._driver_ssh.create_directory, "/path")

        self._driver_ssh._locked_selectfs.assert_called_with(
            *locked_selectfs_args)
        ssh.HNASSSHBackend.check_directory.assert_called_with('/path')
        self.assertTrue(self.mock_log.warning.called)

    def test_create_directory_context_change_success(self):
        locked_selectfs_args = ['create', '/path']
        self.mock_object(time, 'sleep')
        self.mock_object(ssh.HNASSSHBackend, "_locked_selectfs")
        self.mock_object(ssh.HNASSSHBackend, "check_directory",
                         mock.Mock(side_effect=[False, False, True]))

        self._driver_ssh.create_directory("/path")

        self._driver_ssh._locked_selectfs.assert_called_with(
            *locked_selectfs_args)
        ssh.HNASSSHBackend.check_directory.assert_called_with('/path')
        self.assertTrue(self.mock_log.warning.called)

    def test_delete_directory(self):
        locked_selectfs_args = ['delete', '/path']
        self.mock_object(ssh.HNASSSHBackend, "_locked_selectfs")
        self.mock_object(ssh.HNASSSHBackend, "check_directory",
                         mock.Mock(return_value=False))

        self._driver_ssh.delete_directory("/path")

        self._driver_ssh._locked_selectfs.assert_called_with(
            *locked_selectfs_args)
        ssh.HNASSSHBackend.check_directory.assert_called_once_with('/path')
        self.assertFalse(self.mock_log.debug.called)

    def test_delete_directory_directory_not_empty(self):
        locked_selectfs_args = ['delete', '/path']
        self.mock_object(ssh.HNASSSHBackend, "_locked_selectfs", mock.Mock(
            side_effect=exception.HNASDirectoryNotEmpty(msg='fake')))
        self.mock_object(ssh.HNASSSHBackend, "check_directory")

        self._driver_ssh.delete_directory("/path")

        self._driver_ssh._locked_selectfs.assert_called_with(
            *locked_selectfs_args)
        ssh.HNASSSHBackend.check_directory.assert_not_called()
        self.assertFalse(self.mock_log.debug.called)

    def test_delete_directory_context_change_fail(self):
        locked_selectfs_args = ['delete', '/path']
        self.mock_object(time, 'sleep')
        self.mock_object(ssh.HNASSSHBackend, "_locked_selectfs")
        self.mock_object(ssh.HNASSSHBackend, "check_directory",
                         mock.Mock(return_value=True))

        self.assertRaises(exception.HNASSSCContextChange,
                          self._driver_ssh.delete_directory, "/path")

        self._driver_ssh._locked_selectfs.assert_called_with(
            *locked_selectfs_args)
        ssh.HNASSSHBackend.check_directory.assert_called_with('/path')
        self.assertTrue(self.mock_log.debug.called)

    def test_delete_directory_context_change_success(self):
        locked_selectfs_args = ['delete', '/path']
        self.mock_object(time, 'sleep')
        self.mock_object(ssh.HNASSSHBackend, "_locked_selectfs")
        self.mock_object(ssh.HNASSSHBackend, "check_directory",
                         mock.Mock(side_effect=[True, True, False]))

        self._driver_ssh.delete_directory("/path")

        self._driver_ssh._locked_selectfs.assert_called_with(
            *locked_selectfs_args)
        ssh.HNASSSHBackend.check_directory.assert_called_with('/path')
        self.assertTrue(self.mock_log.debug.called)

    def test_check_directory(self):
        path = ("/snapshots/" + self.snapshot['share_id'] + "/" +
                self.snapshot['id'])
        check_snap_args = ['path-to-object-number', '-f', self.fs_name, path]

        self.mock_object(ssh.HNASSSHBackend, '_execute')

        out = self._driver_ssh.check_directory(path)

        self.assertTrue(out)
        self._driver_ssh._execute.assert_called_with(check_snap_args)

    def test_check_directory_retry(self):
        error_msg = ("Unable to run path-to-object-number as "
                     "path-to-object-number is currently running on volume "
                     "39.")
        path = ("/snapshots/" + self.snapshot['share_id'] + "/" +
                self.snapshot['id'])

        check_snap_args = ['path-to-object-number', '-f', self.fs_name, path]

        self.mock_object(time, "sleep")
        self.mock_object(ssh.HNASSSHBackend, '_execute',
                         mock.Mock(side_effect=[putils.ProcessExecutionError(
                             stdout=error_msg), putils.ProcessExecutionError(
                             stdout=error_msg), 'Object number: 0x45a4']))

        out = self._driver_ssh.check_directory(path)

        self.assertIs(True, out)
        self._driver_ssh._execute.assert_called_with(check_snap_args)

    def test_check_inexistent_snapshot(self):
        path = "/path/snap1/snapshot07-08-2016"

        check_snap_args = ['path-to-object-number', '-f', self.fs_name, path]

        self.mock_object(ssh.HNASSSHBackend, '_execute',
                         mock.Mock(side_effect=putils.ProcessExecutionError(
                             stdout=HNAS_RESULT_check_snap_error)))

        out = self._driver_ssh.check_directory(path)

        self.assertFalse(out)
        self._driver_ssh._execute.assert_called_with(check_snap_args)

    def test_check_directory_error(self):
        path = "/path/snap1/snapshot07-08-2016"

        check_snap_args = ['path-to-object-number', '-f', self.fs_name, path]

        self.mock_object(ssh.HNASSSHBackend, '_execute',
                         mock.Mock(side_effect=putils.ProcessExecutionError(
                             stdout="Internal Server Error.")))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.check_directory, path)

        self._driver_ssh._execute.assert_called_with(check_snap_args)

    def test_check_fs_mounted_true(self):
        self.mock_object(ssh.HNASSSHBackend, "_execute",
                         mock.Mock(return_value=(HNAS_RESULT_df, '')))

        self.assertTrue(self._driver_ssh.check_fs_mounted())

    def test_check_fs_mounted_false(self):
        self.mock_object(
            ssh.HNASSSHBackend, "_execute",
            mock.Mock(return_value=(HNAS_RESULT_df_unmounted, '')))

        self.assertFalse(self._driver_ssh.check_fs_mounted())

    def test_check_fs_mounted_error(self):
        self.mock_object(
            ssh.HNASSSHBackend, "_execute",
            mock.Mock(return_value=(HNAS_RESULT_df_error, '')))

        self.assertRaises(exception.HNASItemNotFoundException,
                          self._driver_ssh.check_fs_mounted)

    def test_mount_already_mounted(self):
        fake_mount_command = ['mount', self.fs_name]
        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            side_effect=putils.ProcessExecutionError(stderr='')))

        self.assertRaises(
            exception.HNASBackendException, self._driver_ssh.mount)

        self._driver_ssh._execute.assert_called_with(fake_mount_command)

    def test_vvol_create(self):
        fake_vvol_create_command = ['virtual-volume', 'add', '--ensure',
                                    self.fs_name, 'vvol', '/shares/vvol']
        self.mock_object(ssh.HNASSSHBackend, "_execute")

        self._driver_ssh.vvol_create("vvol")

        self._driver_ssh._execute.assert_called_with(fake_vvol_create_command)

    def test_vvol_create_error(self):
        fake_vvol_create_command = ['virtual-volume', 'add', '--ensure',
                                    self.fs_name, 'vvol', '/shares/vvol']
        self.mock_object(ssh.HNASSSHBackend, "_execute",
                         mock.Mock(side_effect=putils.ProcessExecutionError))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.vvol_create, "vvol")

        self._driver_ssh._execute.assert_called_with(fake_vvol_create_command)

    def test_vvol_delete_vvol_does_not_exist(self):
        fake_vvol_delete_command = ['tree-delete-job-submit', '--confirm',
                                    '-f', self.fs_name, '/shares/vvol']
        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            side_effect=[putils.ProcessExecutionError(
                stderr='Source path: Cannot access')]
        ))

        self._driver_ssh.vvol_delete("vvol")

        self.assertTrue(self.mock_log.warning.called)
        self._driver_ssh._execute.assert_called_with(fake_vvol_delete_command)

    def test_vvol_delete_error(self):
        fake_vvol_delete_command = ['tree-delete-job-submit', '--confirm',
                                    '-f', self.fs_name, '/shares/vvol']
        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            side_effect=[putils.ProcessExecutionError(
                stderr='')]
        ))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.vvol_delete, "vvol")
        self.assertTrue(self.mock_log.exception.called)
        self._driver_ssh._execute.assert_called_with(fake_vvol_delete_command)

    def test_quota_add(self):
        fake_add_quota_command = ['quota', 'add', '--usage-limit', '1G',
                                  '--usage-hard-limit', 'yes',
                                  self.fs_name, 'vvol']
        self.mock_object(ssh.HNASSSHBackend, "_execute")

        self._driver_ssh.quota_add('vvol', 1)

        self._driver_ssh._execute.assert_called_with(fake_add_quota_command)

    def test_modify_quota(self):
        fake_modify_quota_command = ['quota', 'mod', '--usage-limit', '1G',
                                     self.fs_name, 'vvol']
        self.mock_object(ssh.HNASSSHBackend, "_execute")

        self._driver_ssh.modify_quota('vvol', 1)

        self._driver_ssh._execute.assert_called_with(fake_modify_quota_command)

    def test_quota_add_error(self):
        fake_add_quota_command = ['quota', 'add', '--usage-limit', '1G',
                                  '--usage-hard-limit', 'yes',
                                  self.fs_name, 'vvol']
        self.mock_object(ssh.HNASSSHBackend, "_execute",
                         mock.Mock(side_effect=putils.ProcessExecutionError))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.quota_add, 'vvol', 1)

        self._driver_ssh._execute.assert_called_with(fake_add_quota_command)

    def test_modify_quota_error(self):
        fake_modify_quota_command = ['quota', 'mod', '--usage-limit', '1G',
                                     self.fs_name, 'vvol']
        self.mock_object(ssh.HNASSSHBackend, "_execute",
                         mock.Mock(side_effect=putils.ProcessExecutionError))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.modify_quota, 'vvol', 1)

        self._driver_ssh._execute.assert_called_with(fake_modify_quota_command)

    def test_check_vvol(self):
        fake_check_vvol_command = ['virtual-volume', 'list', '--verbose',
                                   self.fs_name, 'vvol']
        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            side_effect=putils.ProcessExecutionError(stderr='')))

        self.assertRaises(exception.HNASItemNotFoundException,
                          self._driver_ssh.check_vvol, 'vvol')
        self._driver_ssh._execute.assert_called_with(fake_check_vvol_command)

    def test_check_quota(self):
        fake_check_quota_command = ['quota', 'list', '--verbose',
                                    self.fs_name, 'vvol']
        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            return_value=('No quotas matching specified filter criteria', '')))

        self.assertRaises(exception.HNASItemNotFoundException,
                          self._driver_ssh.check_quota, 'vvol')
        self._driver_ssh._execute.assert_called_with(fake_check_quota_command)

    def test_check_quota_error(self):
        fake_check_quota_command = ['quota', 'list', '--verbose',
                                    self.fs_name, 'vvol']
        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            side_effect=putils.ProcessExecutionError))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.check_quota, 'vvol')
        self._driver_ssh._execute.assert_called_with(fake_check_quota_command)

    @ddt.data(True, False)
    def test_check_export(self, is_snapshot):
        self.mock_object(ssh.HNASSSHBackend, "_get_export", mock.Mock(
            return_value=[ssh.Export(HNAS_RESULT_export)]))

        self._driver_ssh.check_export("vvol_test", is_snapshot)

    def test_check_export_error(self):
        self.mock_object(ssh.HNASSSHBackend, "_get_export", mock.Mock(
            return_value=[ssh.Export(HNAS_RESULT_wrong_export)]))

        self.assertRaises(exception.HNASItemNotFoundException,
                          self._driver_ssh.check_export, "vvol_test")

    def test_check_cifs(self):
        check_cifs_share_command = ['cifs-share', 'list', 'vvol_test']

        self.mock_object(ssh.HNASSSHBackend, '_execute', mock.Mock(
            return_value=[HNAS_RESULT_cifs_list, '']))

        self._driver_ssh.check_cifs('vvol_test')

        self._driver_ssh._execute.assert_called_with(check_cifs_share_command)

    def test_check_cifs_inexistent_share(self):
        check_cifs_share_command = ['cifs-share', 'list', 'wrong_vvol']

        self.mock_object(ssh.HNASSSHBackend, '_execute', mock.Mock(
            side_effect=[putils.ProcessExecutionError(
                stderr='Export wrong_vvol does not exist on backend '
                       'anymore.')]))

        self.assertRaises(exception.HNASItemNotFoundException,
                          self._driver_ssh.check_cifs, 'wrong_vvol')
        self._driver_ssh._execute.assert_called_with(check_cifs_share_command)

    def test_check_cifs_exception(self):
        check_cifs_share_command = ['cifs-share', 'list', 'wrong_vvol']

        self.mock_object(ssh.HNASSSHBackend, '_execute', mock.Mock(
            side_effect=[putils.ProcessExecutionError(stderr='Error.')]))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.check_cifs, 'wrong_vvol')
        self._driver_ssh._execute.assert_called_with(check_cifs_share_command)

    def test_check_cifs_different_fs_exception(self):
        check_cifs_share_command = ['cifs-share', 'list', 'vvol_test']

        self.mock_object(ssh.HNASSSHBackend, '_execute', mock.Mock(
            return_value=[HNAS_RESULT_different_fs_cifs_list, '']))

        self.assertRaises(exception.HNASItemNotFoundException,
                          self._driver_ssh.check_cifs, 'vvol_test')
        self._driver_ssh._execute.assert_called_with(check_cifs_share_command)

    def test_is_cifs_in_use(self):
        check_cifs_share_command = ['cifs-share', 'list', 'vvol_test']

        self.mock_object(ssh.HNASSSHBackend, '_execute', mock.Mock(
            return_value=[HNAS_RESULT_cifs_list, '']))

        out = self._driver_ssh.is_cifs_in_use('vvol_test')

        self.assertTrue(out)
        self._driver_ssh._execute.assert_called_with(check_cifs_share_command)

    def test_is_cifs_without_use(self):
        check_cifs_share_command = ['cifs-share', 'list', 'vvol_test']

        self.mock_object(ssh.HNASSSHBackend, '_execute', mock.Mock(
            return_value=[HNAS_RESULT_different_fs_cifs_list, '']))

        out = self._driver_ssh.is_cifs_in_use('vvol_test')

        self.assertFalse(out)
        self._driver_ssh._execute.assert_called_with(check_cifs_share_command)

    def test_get_share_quota(self):
        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            return_value=(HNAS_RESULT_quota, '')))

        result = self._driver_ssh.get_share_quota("vvol_test")

        self.assertEqual(5, result)

    @ddt.data(HNAS_RESULT_quota_unset, HNAS_RESULT_quota_err)
    def test_get_share_quota_errors(self, hnas_output):
        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            return_value=(hnas_output, '')))

        result = self._driver_ssh.get_share_quota("vvol_test")

        self.assertIsNone(result)

    def test_get_share_quota_tb(self):
        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            return_value=(HNAS_RESULT_quota_tb, '')))

        result = self._driver_ssh.get_share_quota("vvol_test")

        self.assertEqual(1024, result)

    def test_get_share_quota_mb(self):
        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            return_value=(HNAS_RESULT_quota_mb, '')))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh.get_share_quota, "vvol_test")

    def test_get_share_usage(self):
        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            return_value=(HNAS_RESULT_quota, '')))

        self.assertEqual(1, self._driver_ssh.get_share_usage("vvol_test"))

    def test_get_share_usage_error(self):
        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            return_value=(HNAS_RESULT_quota_err, '')))

        self.assertRaises(exception.HNASItemNotFoundException,
                          self._driver_ssh.get_share_usage, "vvol_test")

    def test_get_share_usage_mb(self):
        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            return_value=(HNAS_RESULT_quota_mb, '')))

        self.assertEqual(0.01953125, self._driver_ssh.get_share_usage(
            "vvol_test"))

    def test_get_share_usage_tb(self):
        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            return_value=(HNAS_RESULT_quota_tb, '')))

        self.assertEqual(1024, self._driver_ssh.get_share_usage("vvol_test"))

    @ddt.data(True, False)
    def test__get_share_export(self, is_snapshot):
        self.mock_object(ssh.HNASSSHBackend, '_execute', mock.Mock(
            return_value=[HNAS_RESULT_export_ip, '']))

        export_list = self._driver_ssh._get_export(
            name='fake_name', is_snapshot=is_snapshot)
        path = '/shares/fake_name'
        if is_snapshot:
            path = '/snapshots/fake_name'

        command = ['nfs-export', 'list ', path]

        self._driver_ssh._execute.assert_called_with(command)
        self.assertEqual('vvol_test', export_list[0].export_name)
        self.assertEqual('/vvol_test', export_list[0].export_path)
        self.assertEqual('fake_fs', export_list[0].file_system_label)
        self.assertEqual('Yes', export_list[0].mounted)
        self.assertIn('rw', export_list[0].export_configuration[0])

    def test__get_share_export_fs_not_available(self):

        self.mock_object(ssh.HNASSSHBackend, '_execute', mock.Mock(
            return_value=[HNAS_RESULT_exp_no_fs, '']))

        export_list = self._driver_ssh._get_export(name='fake_name')
        path = '/shares/fake_name'

        command = ['nfs-export', 'list ', path]

        self._driver_ssh._execute.assert_called_with(command)
        self.assertEqual('no_fs', export_list[0].export_name)
        self.assertEqual('/export_without_fs', export_list[0].export_path)
        self.assertEqual('*** not available ***',
                         export_list[0].file_system_info)
        self.assertEqual([], export_list[0].export_configuration)
        not_in_keys = ['file_system_label', 'file_system_size', 'formatted',
                       'file_system_free_space', 'file_system_state', 'failed',
                       'mounted', 'thin_provisioned']
        for key in not_in_keys:
            self.assertNotIn(key, export_list[0].__dict__)

    def test__get_share_export_exception_not_found(self):

        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            side_effect=putils.ProcessExecutionError(
                stderr="NFS Export List: Export 'id' does not exist.")))

        self.assertRaises(exception.HNASItemNotFoundException,
                          self._driver_ssh._get_export, 'fake_id')

    def test__get_share_export_exception_error(self):

        self.mock_object(ssh.HNASSSHBackend, "_execute", mock.Mock(
            side_effect=putils.ProcessExecutionError(stderr="Some error.")
        ))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh._get_export, 'fake_id')

    def test__execute(self):
        key = self.ssh_private_key
        commands = ['tree-clone-job-submit', '-e', '/src', '/dst']
        concat_command = ('ssc --smuauth fake console-context --evs 2 '
                          'tree-clone-job-submit -e /src /dst')
        self.mock_object(paramiko.SSHClient, 'connect')
        self.mock_object(putils, 'ssh_execute',
                         mock.Mock(return_value=[HNAS_RESULT_job, '']))

        output, err = self._driver_ssh._execute(commands)

        putils.ssh_execute.assert_called_once_with(mock.ANY, concat_command,
                                                   check_exit_code=True)
        paramiko.SSHClient.connect.assert_called_with(self.ip,
                                                      username=self.user,
                                                      key_filename=key,
                                                      look_for_keys=False,
                                                      timeout=None,
                                                      password=self.password,
                                                      port=self.port,
                                                      banner_timeout=None)
        self.assertIn('Request submitted successfully.', output)

    def test__execute_ssh_exception(self):
        commands = ['tree-clone-job-submit', '-e', '/src', '/dst']
        concat_command = ('ssc --smuauth fake console-context --evs 2 '
                          'tree-clone-job-submit -e /src /dst')
        msg = 'Failed to establish SSC connection'

        self.mock_object(time, "sleep")
        self.mock_object(paramiko.SSHClient, 'connect')
        self.mock_object(putils, 'ssh_execute',
                         mock.Mock(side_effect=[
                             putils.ProcessExecutionError(stderr=msg),
                             putils.ProcessExecutionError(stderr='Invalid!')]))
        self.mock_object(mutils.SSHPool, "item",
                         mock.Mock(return_value=paramiko.SSHClient()))
        self.mock_object(paramiko.SSHClient, "set_missing_host_key_policy")

        self.assertRaises(putils.ProcessExecutionError,
                          self._driver_ssh._execute, commands)

        putils.ssh_execute.assert_called_with(mock.ANY, concat_command,
                                              check_exit_code=True)

        self.assertTrue(self.mock_log.debug.called)

    def test__locked_selectfs_create_operation(self):
        exec_command = ['selectfs', self.fs_name, '\n', 'ssc', '127.0.0.1',
                        'console-context', '--evs', six.text_type(self.evs_id),
                        'mkdir', '-p', '/path']
        self.mock_object(ssh.HNASSSHBackend, '_execute')

        self._driver_ssh._locked_selectfs('create', '/path')

        self._driver_ssh._execute.assert_called_with(exec_command)

    def test__locked_selectfs_create_operation_error(self):
        exec_command = ['selectfs', self.fs_name, '\n', 'ssc', '127.0.0.1',
                        'console-context', '--evs', six.text_type(self.evs_id),
                        'mkdir', '-p', '/path']
        self.mock_object(
            ssh.HNASSSHBackend, '_execute',
            mock.Mock(side_effect=putils.ProcessExecutionError(
                stderr="some error")))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh._locked_selectfs, 'create', '/path')

        self._driver_ssh._execute.assert_called_with(exec_command)

    def test__locked_selectfs_create_operation_context_change(self):
        exec_command = ['selectfs', self.fs_name, '\n', 'ssc', '127.0.0.1',
                        'console-context', '--evs', six.text_type(self.evs_id),
                        'mkdir', '-p', '/path']
        self.mock_object(
            ssh.HNASSSHBackend, '_execute',
            mock.Mock(side_effect=putils.ProcessExecutionError(
                stderr="Current file system invalid: VolumeNotFound")))

        self.assertRaises(exception.HNASSSCContextChange,
                          self._driver_ssh._locked_selectfs, 'create', '/path')

        self._driver_ssh._execute.assert_called_with(exec_command)
        self.assertTrue(self.mock_log.debug.called)

    def test__locked_selectfs_delete_operation_successful(self):
        exec_command = ['selectfs', self.fs_name, '\n', 'ssc', '127.0.0.1',
                        'console-context', '--evs', six.text_type(self.evs_id),
                        'rmdir', '/path']
        self.mock_object(ssh.HNASSSHBackend, '_execute')

        self._driver_ssh._locked_selectfs('delete', '/path')

        self._driver_ssh._execute.assert_called_with(exec_command)

    def test__locked_selectfs_deleting_not_empty_directory(self):
        msg = 'This path has more snapshot. Currenty DirectoryNotEmpty'

        self.mock_object(ssh.HNASSSHBackend, '_execute', mock.Mock(
            side_effect=[putils.ProcessExecutionError(stderr=msg)]))

        self.assertRaises(exception.HNASDirectoryNotEmpty,
                          self._driver_ssh._locked_selectfs, 'delete', '/path')

        self.assertTrue(self.mock_log.debug.called)

    def test__locked_selectfs_delete_exception(self):
        msg = "rmdir: cannot remove '/path'"

        self.mock_object(ssh.HNASSSHBackend, '_execute', mock.Mock(
            side_effect=[putils.ProcessExecutionError(stderr=msg)]))

        self.assertRaises(exception.HNASBackendException,
                          self._driver_ssh._locked_selectfs, 'delete', 'path')
        self.assertTrue(self.mock_log.exception.called)

    def test__locked_selectfs_delete_not_found(self):
        msg = "rmdir: cannot remove '/path': NotFound"

        self.mock_object(ssh.HNASSSHBackend, '_execute', mock.Mock(
            side_effect=[putils.ProcessExecutionError(stderr=msg)]))

        self._driver_ssh._locked_selectfs('delete', 'path')

        self.assertTrue(self.mock_log.warning.called)

    def test__locked_selectfs_delete_context_change(self):
        msg = "Current file system invalid: VolumeNotFound"

        self.mock_object(ssh.HNASSSHBackend, '_execute', mock.Mock(
            side_effect=[putils.ProcessExecutionError(stderr=msg)]))

        self.assertRaises(exception.HNASSSCContextChange,
                          self._driver_ssh._locked_selectfs, 'delete', 'path')

        self.assertTrue(self.mock_log.debug.called)



# Copyright 2015 Mirantis Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add more network info attributes to 'network_allocations' table.

Revision ID: 5155c7077f99
Revises: 293fac1130ca
Create Date: 2015-12-22 12:05:24.297049

"""

# revision identifiers, used by Alembic.
revision = '5155c7077f99'
down_revision = '293fac1130ca'

from alembic import op
import sqlalchemy as sa


def upgrade():
    default_label_value = 'user'
    op.add_column(
        'network_allocations',
        sa.Column('label',
                  sa.String(255),
                  default=default_label_value,
                  server_default=default_label_value,
                  nullable=True),
    )
    op.add_column(
        'network_allocations',
        sa.Column('network_type', sa.String(32), nullable=True))
    op.add_column(
        'network_allocations',
        sa.Column('segmentation_id', sa.Integer, nullable=True))
    op.add_column(
        'network_allocations',
        sa.Column('ip_version', sa.Integer, nullable=True))
    op.add_column(
        'network_allocations',
        sa.Column('cidr', sa.String(64), nullable=True))


def downgrade():
    for col_name in ('label', 'network_type', 'segmentation_id', 'ip_version',
                     'cidr'):
        op.drop_column('network_allocations', col_name)



# Copyright (c) 2014 Alex Meade.  All rights reserved.
# Copyright (c) 2014 Clinton Knight.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log
from oslo_utils import excutils

from manila.share.drivers.netapp.dataontap.client import api as netapp_api
from manila.share.drivers.netapp import utils as na_utils


LOG = log.getLogger(__name__)


class NetAppBaseClient(object):

    def __init__(self, **kwargs):
        self.connection = netapp_api.NaServer(
            host=kwargs['hostname'],
            transport_type=kwargs['transport_type'],
            port=kwargs['port'],
            username=kwargs['username'],
            password=kwargs['password'],
            trace=kwargs.get('trace', False),
            api_trace_pattern=kwargs.get('api_trace_pattern',
                                         na_utils.API_TRACE_PATTERN))

    def get_ontapi_version(self, cached=True):
        """Gets the supported ontapi version."""

        if cached:
            return self.connection.get_api_version()

        result = self.send_request('system-get-ontapi-version',
                                   enable_tunneling=False)
        major = result.get_child_content('major-version')
        minor = result.get_child_content('minor-version')
        return major, minor

    @na_utils.trace
    def get_system_version(self, cached=True):
        """Gets the current Data ONTAP version."""

        if cached:
            return self.connection.get_system_version()

        result = self.send_request('system-get-version')

        version_tuple = result.get_child_by_name(
            'version-tuple') or netapp_api.NaElement('none')
        system_version_tuple = version_tuple.get_child_by_name(
            'system-version-tuple') or netapp_api.NaElement('none')

        version = {}
        version['version'] = result.get_child_content('version')
        version['version-tuple'] = (
            int(system_version_tuple.get_child_content('generation')),
            int(system_version_tuple.get_child_content('major')),
            int(system_version_tuple.get_child_content('minor')))

        return version

    def _init_features(self):
        """Set up the repository of available Data ONTAP features."""
        self.features = Features()

    def _strip_xml_namespace(self, string):
        if string.startswith('{') and '}' in string:
            return string.split('}', 1)[1]
        return string

    def send_request(self, api_name, api_args=None, enable_tunneling=True):
        """Sends request to Ontapi."""
        request = netapp_api.NaElement(api_name)
        if api_args:
            request.translate_struct(api_args)
        return self.connection.invoke_successfully(request, enable_tunneling)

    @na_utils.trace
    def get_licenses(self):
        try:
            result = self.send_request('license-v2-list-info')
        except netapp_api.NaApiError:
            with excutils.save_and_reraise_exception():
                LOG.exception("Could not get licenses list.")

        return sorted(
            [l.get_child_content('package').lower()
             for l in result.get_child_by_name('licenses').get_children()])

    def send_ems_log_message(self, message_dict):
        """Sends a message to the Data ONTAP EMS log."""
        raise NotImplementedError()


class Features(object):

    def __init__(self):
        self.defined_features = set()

    def add_feature(self, name, supported=True):
        if not isinstance(supported, bool):
            raise TypeError("Feature value must be a bool type.")
        self.defined_features.add(name)
        setattr(self, name, supported)

    def __getattr__(self, name):
        # NOTE(cknight): Needed to keep pylint happy.
        raise AttributeError



# Copyright (c) 2016 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import socket
import sys

from oslo_config import cfg
from oslo_log import log
from oslo_utils import units

from manila.common import constants
from manila import exception
from manila.i18n import _
from manila.share import driver
from manila.share.drivers import ganesha
from manila.share.drivers.ganesha import utils as ganesha_utils
from manila.share import share_types

try:
    import ceph_volume_client
    ceph_module_found = True
except ImportError as e:
    ceph_volume_client = None
    ceph_module_found = False


CEPHX_ACCESS_TYPE = "cephx"

# The default Ceph administrative identity
CEPH_DEFAULT_AUTH_ID = "admin"

DEFAULT_VOLUME_MODE = '755'


LOG = log.getLogger(__name__)

cephfs_opts = [
    cfg.StrOpt('cephfs_conf_path',
               default="",
               help="Fully qualified path to the ceph.conf file."),
    cfg.StrOpt('cephfs_cluster_name',
               help="The name of the cluster in use, if it is not "
                    "the default ('ceph')."
               ),
    cfg.StrOpt('cephfs_auth_id',
               default="manila",
               help="The name of the ceph auth identity to use."
               ),
    cfg.StrOpt('cephfs_volume_path_prefix',
               default="/volumes",
               help="The prefix of the cephfs volume path."
               ),
    cfg.BoolOpt('cephfs_enable_snapshots',
                default=False,
                help="Whether to enable snapshots in this driver."
                ),
    cfg.StrOpt('cephfs_protocol_helper_type',
               default="CEPHFS",
               choices=['CEPHFS', 'NFS'],
               ignore_case=True,
               help="The type of protocol helper to use. Default is "
                    "CEPHFS."
               ),
    cfg.BoolOpt('cephfs_ganesha_server_is_remote',
                default=False,
                help="Whether the NFS-Ganesha server is remote to the driver."
                ),
    cfg.StrOpt('cephfs_ganesha_server_ip',
               help="The IP address of the NFS-Ganesha server."),
    cfg.StrOpt('cephfs_ganesha_server_username',
               default='root',
               help="The username to authenticate as in the remote "
                    "NFS-Ganesha server host."),
    cfg.StrOpt('cephfs_ganesha_path_to_private_key',
               help="The path of the driver host's private SSH key file."),
    cfg.StrOpt('cephfs_ganesha_server_password',
               secret=True,
               help="The password to authenticate as the user in the remote "
                    "Ganesha server host. This is not required if "
                    "'cephfs_ganesha_path_to_private_key' is configured."),
    cfg.StrOpt('cephfs_volume_mode',
               default=DEFAULT_VOLUME_MODE,
               help="The read/write/execute permissions mode for CephFS "
                    "volumes, snapshots, and snapshot groups expressed in "
                    "Octal as with linux 'chmod' or 'umask' commands."),
]


CONF = cfg.CONF
CONF.register_opts(cephfs_opts)


def cephfs_share_path(share):
    """Get VolumePath from Share."""
    return ceph_volume_client.VolumePath(
        share['share_group_id'], share['id'])


class CephFSDriver(driver.ExecuteMixin, driver.GaneshaMixin,
                   driver.ShareDriver,):
    """Driver for the Ceph Filesystem."""

    def __init__(self, *args, **kwargs):
        super(CephFSDriver, self).__init__(False, *args, **kwargs)
        self.backend_name = self.configuration.safe_get(
            'share_backend_name') or 'CephFS'

        self._volume_client = None

        self.configuration.append_config_values(cephfs_opts)

        try:
            self._cephfs_volume_mode = int(
                self.configuration.cephfs_volume_mode, 8)
        except ValueError:
            msg = _("Invalid CephFS volume mode %s")
            raise exception.BadConfigurationException(
                msg % self.configuration.cephfs_volume_mode)

    def do_setup(self, context):
        if self.configuration.cephfs_protocol_helper_type.upper() == "CEPHFS":
            protocol_helper_class = getattr(
                sys.modules[__name__], 'NativeProtocolHelper')
        else:
            protocol_helper_class = getattr(
                sys.modules[__name__], 'NFSProtocolHelper')

        self.protocol_helper = protocol_helper_class(
            self._execute,
            self.configuration,
            ceph_vol_client=self.volume_client)

        self.protocol_helper.init_helper()

    def _update_share_stats(self):
        stats = self.volume_client.rados.get_cluster_stats()

        total_capacity_gb = stats['kb'] * units.Mi
        free_capacity_gb = stats['kb_avail'] * units.Mi

        data = {
            'vendor_name': 'Ceph',
            'driver_version': '1.0',
            'share_backend_name': self.backend_name,
            'storage_protocol': self.configuration.safe_get(
                'cephfs_protocol_helper_type'),
            'pools': [
                {
                    'pool_name': 'cephfs',
                    'total_capacity_gb': total_capacity_gb,
                    'free_capacity_gb': free_capacity_gb,
                    'qos': 'False',
                    'reserved_percentage': 0,
                    'dedupe': [False],
                    'compression': [False],
                    'thin_provisioning': [False]
                }
            ],
            'total_capacity_gb': total_capacity_gb,
            'free_capacity_gb': free_capacity_gb,
            'snapshot_support': self.configuration.safe_get(
                'cephfs_enable_snapshots'),
        }
        super(CephFSDriver, self)._update_share_stats(data)

    def _to_bytes(self, gigs):
        """Convert a Manila size into bytes.

        Manila uses gibibytes everywhere.

        :param gigs: integer number of gibibytes.
        :return: integer number of bytes.
        """
        return gigs * units.Gi

    @property
    def volume_client(self):
        if self._volume_client:
            return self._volume_client

        if not ceph_module_found:
            raise exception.ManilaException(
                _("Ceph client libraries not found.")
            )

        conf_path = self.configuration.safe_get('cephfs_conf_path')
        cluster_name = self.configuration.safe_get('cephfs_cluster_name')
        auth_id = self.configuration.safe_get('cephfs_auth_id')
        volume_prefix = self.configuration.safe_get(
            'cephfs_volume_path_prefix')
        self._volume_client = ceph_volume_client.CephFSVolumeClient(
            auth_id, conf_path, cluster_name, volume_prefix=volume_prefix)
        LOG.info("[%(be)s}] Ceph client found, connecting...",
                 {"be": self.backend_name})
        if auth_id != CEPH_DEFAULT_AUTH_ID:
            # Evict any other manila sessions.  Only do this if we're
            # using a client ID that isn't the default admin ID, to avoid
            # rudely disrupting anyone else.
            premount_evict = auth_id
        else:
            premount_evict = None
        try:
            self._volume_client.connect(premount_evict=premount_evict)
        except Exception:
            self._volume_client = None
            raise
        else:
            LOG.info("[%(be)s] Ceph client connection complete.",
                     {"be": self.backend_name})

        return self._volume_client

    def create_share(self, context, share, share_server=None):
        """Create a CephFS volume.

        :param context: A RequestContext.
        :param share: A Share.
        :param share_server: Always None for CephFS native.
        :return: The export locations dictionary.
        """
        requested_proto = share['share_proto'].upper()
        supported_proto = (
            self.configuration.cephfs_protocol_helper_type.upper())
        if (requested_proto != supported_proto):
            msg = _("Share protocol %s is not supported.") % requested_proto
            raise exception.ShareBackendException(msg=msg)

        # `share` is a Share
        msg = _("create_share {be} name={id} size={size}"
                " share_group_id={group}")
        LOG.debug(msg.format(
            be=self.backend_name, id=share['id'], size=share['size'],
            group=share['share_group_id']))

        extra_specs = share_types.get_extra_specs_from_share(share)
        data_isolated = extra_specs.get("cephfs:data_isolated", False)

        size = self._to_bytes(share['size'])

        # Create the CephFS volume
        cephfs_volume = self.volume_client.create_volume(
            cephfs_share_path(share), size=size, data_isolated=data_isolated,
            mode=self._cephfs_volume_mode)

        return self.protocol_helper.get_export_locations(share, cephfs_volume)

    def delete_share(self, context, share, share_server=None):
        extra_specs = share_types.get_extra_specs_from_share(share)
        data_isolated = extra_specs.get("cephfs:data_isolated", False)

        self.volume_client.delete_volume(cephfs_share_path(share),
                                         data_isolated=data_isolated)
        self.volume_client.purge_volume(cephfs_share_path(share),
                                        data_isolated=data_isolated)

    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        return self.protocol_helper.update_access(
            context, share, access_rules, add_rules, delete_rules,
            share_server=share_server)

    def ensure_share(self, context, share, share_server=None):
        # Creation is idempotent
        return self.create_share(context, share, share_server)

    def extend_share(self, share, new_size, share_server=None):
        LOG.debug("extend_share {id} {size}".format(
            id=share['id'], size=new_size))
        self.volume_client.set_max_bytes(cephfs_share_path(share),
                                         self._to_bytes(new_size))

    def shrink_share(self, share, new_size, share_server=None):
        LOG.debug("shrink_share {id} {size}".format(
            id=share['id'], size=new_size))
        new_bytes = self._to_bytes(new_size)
        used = self.volume_client.get_used_bytes(cephfs_share_path(share))
        if used > new_bytes:
            # While in fact we can "shrink" our volumes to less than their
            # used bytes (it's just a quota), raise error anyway to avoid
            # confusing API consumers that might depend on typical shrink
            # behaviour.
            raise exception.ShareShrinkingPossibleDataLoss(
                share_id=share['id'])

        self.volume_client.set_max_bytes(cephfs_share_path(share), new_bytes)

    def create_snapshot(self, context, snapshot, share_server=None):
        self.volume_client.create_snapshot_volume(
            cephfs_share_path(snapshot['share']),
            '_'.join([snapshot['snapshot_id'], snapshot['id']]),
            mode=self._cephfs_volume_mode)

    def delete_snapshot(self, context, snapshot, share_server=None):
        self.volume_client.destroy_snapshot_volume(
            cephfs_share_path(snapshot['share']),
            '_'.join([snapshot['snapshot_id'], snapshot['id']]))

    def create_share_group(self, context, sg_dict, share_server=None):
        self.volume_client.create_group(sg_dict['id'],
                                        mode=self._cephfs_volume_mode)

    def delete_share_group(self, context, sg_dict, share_server=None):
        self.volume_client.destroy_group(sg_dict['id'])

    def delete_share_group_snapshot(self, context, snap_dict,
                                    share_server=None):
        self.volume_client.destroy_snapshot_group(
            snap_dict['share_group_id'],
            snap_dict['id'])

        return None, []

    def create_share_group_snapshot(self, context, snap_dict,
                                    share_server=None):
        self.volume_client.create_snapshot_group(
            snap_dict['share_group_id'],
            snap_dict['id'],
            mode=self._cephfs_volume_mode)

        return None, []

    def __del__(self):
        if self._volume_client:
            self._volume_client.disconnect()
            self._volume_client = None


class NativeProtocolHelper(ganesha.NASHelperBase):
    """Helper class for native CephFS protocol"""

    supported_access_types = (CEPHX_ACCESS_TYPE, )
    supported_access_levels = (constants.ACCESS_LEVEL_RW,
                               constants.ACCESS_LEVEL_RO)

    def __init__(self, execute, config, **kwargs):
        self.volume_client = kwargs.pop('ceph_vol_client')
        super(NativeProtocolHelper, self).__init__(execute, config,
                                                   **kwargs)

    def _init_helper(self):
        pass

    def get_export_locations(self, share, cephfs_volume):
        # To mount this you need to know the mon IPs and the path to the volume
        mon_addrs = self.volume_client.get_mon_addrs()

        export_location = "{addrs}:{path}".format(
            addrs=",".join(mon_addrs),
            path=cephfs_volume['mount_path'])

        LOG.info("Calculated export location for share %(id)s: %(loc)s",
                 {"id": share['id'], "loc": export_location})

        return {
            'path': export_location,
            'is_admin_only': False,
            'metadata': {},
        }

    def _allow_access(self, context, share, access, share_server=None):
        if access['access_type'] != CEPHX_ACCESS_TYPE:
            raise exception.InvalidShareAccess(
                reason=_("Only 'cephx' access type allowed."))

        ceph_auth_id = access['access_to']

        # We need to check here rather than the API or Manila Client to see
        # if the ceph_auth_id is the same as the one specified for Manila's
        # usage. This is due to the fact that the API and the Manila client
        # cannot read the contents of the Manila configuration file. If it
        # is the same, we need to error out.
        if ceph_auth_id == CONF.cephfs_auth_id:
            error_message = (_('Ceph authentication ID %s must be different '
                             'than the one the Manila service uses.') %
                             ceph_auth_id)
            raise exception.InvalidInput(message=error_message)

        if not getattr(self.volume_client, 'version', None):
            if access['access_level'] == constants.ACCESS_LEVEL_RO:
                LOG.error("Need python-cephfs package version 10.2.3 or "
                          "greater to enable read-only access.")
                raise exception.InvalidShareAccessLevel(
                    level=constants.ACCESS_LEVEL_RO)

            auth_result = self.volume_client.authorize(
                cephfs_share_path(share), ceph_auth_id)
        else:
            readonly = access['access_level'] == constants.ACCESS_LEVEL_RO
            auth_result = self.volume_client.authorize(
                cephfs_share_path(share), ceph_auth_id, readonly=readonly,
                tenant_id=share['project_id'])

        return auth_result['auth_key']

    def _deny_access(self, context, share, access, share_server=None):
        if access['access_type'] != CEPHX_ACCESS_TYPE:
            LOG.warning("Invalid access type '%(type)s', "
                        "ignoring in deny.",
                        {"type": access['access_type']})
            return

        self.volume_client.deauthorize(cephfs_share_path(share),
                                       access['access_to'])
        self.volume_client.evict(
            access['access_to'],
            volume_path=cephfs_share_path(share))

    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        access_keys = {}

        if not (add_rules or delete_rules):  # recovery/maintenance mode
            add_rules = access_rules

            existing_auths = None

            # The unversioned volume client cannot fetch from the Ceph backend,
            # the list of auth IDs that have share access.
            if getattr(self.volume_client, 'version', None):
                existing_auths = self.volume_client.get_authorized_ids(
                    cephfs_share_path(share))

            if existing_auths:
                existing_auth_ids = set(
                    [auth[0] for auth in existing_auths])
                want_auth_ids = set(
                    [rule['access_to'] for rule in add_rules])
                delete_auth_ids = existing_auth_ids.difference(
                    want_auth_ids)
                for delete_auth_id in delete_auth_ids:
                    delete_rules.append(
                        {
                            'access_to': delete_auth_id,
                            'access_type': CEPHX_ACCESS_TYPE,
                        })

        # During recovery mode, re-authorize share access for auth IDs that
        # were already granted access by the backend. Do this to fetch their
        # access keys and ensure that after recovery, manila and the Ceph
        # backend are in sync.
        for rule in add_rules:
            access_key = self._allow_access(context, share, rule)
            access_keys.update({rule['access_id']: {'access_key': access_key}})

        for rule in delete_rules:
            self._deny_access(context, share, rule)

        return access_keys


class NFSProtocolHelper(ganesha.GaneshaNASHelper2):

    shared_data = {}
    supported_protocols = ('NFS',)

    def __init__(self, execute, config_object, **kwargs):
        if config_object.cephfs_ganesha_server_is_remote:
            execute = ganesha_utils.SSHExecutor(
                config_object.cephfs_ganesha_server_ip, 22, None,
                config_object.cephfs_ganesha_server_username,
                password=config_object.cephfs_ganesha_server_password,
                privatekey=config_object.cephfs_ganesha_path_to_private_key)
        else:
            execute = ganesha_utils.RootExecutor(execute)

        self.ganesha_host = config_object.cephfs_ganesha_server_ip
        if not self.ganesha_host:
            self.ganesha_host = socket.gethostname()
            LOG.info("NFS-Ganesha server's location defaulted to driver's "
                     "hostname: %s", self.ganesha_host)

        super(NFSProtocolHelper, self).__init__(execute, config_object,
                                                **kwargs)

        if not hasattr(self, 'ceph_vol_client'):
            self.ceph_vol_client = kwargs.pop('ceph_vol_client')

    def get_export_locations(self, share, cephfs_volume):
        export_location = "{server_address}:{path}".format(
            server_address=self.ganesha_host,
            path=cephfs_volume['mount_path'])

        LOG.info("Calculated export location for share %(id)s: %(loc)s",
                 {"id": share['id'], "loc": export_location})

        return {
            'path': export_location,
            'is_admin_only': False,
            'metadata': {},
        }

    def _default_config_hook(self):
        """Callback to provide default export block."""
        dconf = super(NFSProtocolHelper, self)._default_config_hook()
        conf_dir = ganesha_utils.path_from(__file__, "conf")
        ganesha_utils.patch(dconf, self._load_conf_dir(conf_dir))
        return dconf

    def _fsal_hook(self, base, share, access):
        """Callback to create FSAL subblock."""
        ceph_auth_id = ''.join(['ganesha-', share['id']])
        auth_result = self.ceph_vol_client.authorize(
            cephfs_share_path(share), ceph_auth_id, readonly=False,
            tenant_id=share['project_id'])
        # Restrict Ganesha server's access to only the CephFS subtree or path,
        # corresponding to the manila share, that is to be exported by making
        # Ganesha use Ceph auth IDs with path restricted capabilities to
        # communicate with CephFS.
        return {
            'Name': 'Ceph',
            'User_Id': ceph_auth_id,
            'Secret_Access_Key': auth_result['auth_key']
        }

    def _cleanup_fsal_hook(self, base, share, access):
        """Callback for FSAL specific cleanup after removing an export."""
        ceph_auth_id = ''.join(['ganesha-', share['id']])
        self.ceph_vol_client.deauthorize(cephfs_share_path(share),
                                         ceph_auth_id)

    def _get_export_path(self, share):
        """Callback to provide export path."""
        volume_path = cephfs_share_path(share)
        return self.ceph_vol_client._get_path(volume_path)

    def _get_export_pseudo_path(self, share):
        """Callback to provide pseudo path."""
        volume_path = cephfs_share_path(share)
        return self.ceph_vol_client._get_path(volume_path)



# Copyright (c) 2014 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""EMC Share Driver Base Plugin API """

import abc

import six


@six.add_metaclass(abc.ABCMeta)
class StorageConnection(object):
    """Subclasses should implement storage backend specific functionality."""

    def __init__(self, *args, **kwargs):
        # NOTE(vponomaryov): redefine 'driver_handles_share_servers' within
        #                    plugin.
        self.driver_handles_share_servers = None

    @abc.abstractmethod
    def create_share(self, context, share, share_server):
        """Is called to create share."""

    @abc.abstractmethod
    def create_snapshot(self, context, snapshot, share_server):
        """Is called to create snapshot."""

    @abc.abstractmethod
    def delete_share(self, context, share, share_server):
        """Is called to remove share."""

    @abc.abstractmethod
    def delete_snapshot(self, context, snapshot, share_server):
        """Is called to remove snapshot."""

    @abc.abstractmethod
    def ensure_share(self, context, share, share_server):
        """Invoked to ensure that share is exported."""

    @abc.abstractmethod
    def extend_share(self, share, new_size, share_server):
        """Invoked to extend share."""

    @abc.abstractmethod
    def allow_access(self, context, share, access, share_server):
        """Allow access to the share."""

    @abc.abstractmethod
    def deny_access(self, context, share, access, share_server):
        """Deny access to the share."""

    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        """Update access rules for given share."""
        raise NotImplementedError()

    def raise_connect_error(self):
        """Check for setup error."""
        pass

    def connect(self, emc_share_driver, context):
        """Any initialization the share driver does while starting."""
        pass

    def update_share_stats(self, stats_dict):
        """Add key/values to stats_dict."""
        pass

    def get_network_allocations_number(self):
        """Returns number of network allocations for creating VIFs."""
        return 0

    @abc.abstractmethod
    def setup_server(self, network_info, metadata=None):
        """Set up and configure share server with given network parameters."""

    @abc.abstractmethod
    def teardown_server(self, server_details, security_services=None):
        """Teardown share server."""
